--[[
-- Copyright 2017 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]

--[[
-- This module defines some default AI's, and any other helper functions that
-- may be needed to allow game developers to craft their own custom AI.
-- The two most important functions are `dumb` and `average`, each of which is
-- an AI of dumb and average intelligence (smart AI's need to be
-- character-specific).
--
-- A key notion of this AI is that it is as deterministic as possible. The
-- only thing that is *not* deterministic is power selection in the average
-- AI, and that's because powers have such a huge variety, and may be so
-- specific to a particular build, that we can't write a general strategy for
-- selecting them deterministically, at least not one that doesn't just spam
-- the same power over and over again.
-- ]]

local M = {}
local attack = require 'combat/attack'
local combatstateops = require 'combat/combatstate'
local log = require 'log'
local pages =  require 'pages'
local unpack = require 'scarletfunctions'.unpack


--[[
-- A combat page to be used by computer-controlled characters that select
-- their actions.
--]]
M.aichoice = pages.page(
    nil,
    function(combatstate)
        local activecharacter = combatstate.characters[combatstate.charindex]
        if combatstateops.isdefeated(combatstate, activecharacter.name) then
            local name = activecharacter.name
            log.info(activecharacter.combatname .. " defeated. Skipping.")
            return pages.page(
                nil,
                nil,
                nil,
                function(combatstate1)
                    local choices = combatstate1.choicecounts[name]
                    for _, count in pairs(choices) do
                        count.current = 0
                    end
                end
            )
        end
        local ai
        if combatstate.ai == "dumb" then
            log.info("Using the dumb AI for " .. activecharacter.name)
            ai = M.dumb
        elseif combatstate.ai == "smart" and activecharacter.ai then
            log.info("Using the smart AI for " .. activecharacter.name)
            ai = activecharacter.ai
        else
            log.info("Using the average AI for " .. activecharacter.name)
            ai = M.average
        end
        return ai(combatstate)
    end
)

--[[
-- The dumb AI never uses powers. It alternates between attacking the target with the highest 
-- energy and the target with the lowest energy as many times as it can
-- It then selects the target with the best buff and spanks them as many times as it can.
--
-- combatstate - The current state of combat
--
-- return The combat action the active characters should execute, or a page
-- that does nothing but print that the character can't attack, if the
-- character can't attack
-- ]]
function M.dumb(combatstate)
    local character = combatstate.characters[combatstate.charindex]
    log.info(character.name .. " is using the dumb AI.")
    local attackcount = combatstate.choicecounts[character.name].attack
    local shootcount = combatstate.choicecounts[character.name].shoot
    local canattack = attackcount and attackcount.current > 0
    local canshoot = shootcount and shootcount.current > 0
    if not canattack and not canshoot then
        log.info(character.name .. " has finished attacking.")
        return pages.page(
            nil,
            nil,
            nil,
            function(combatstate1)
                local choices = combatstate1.choicecounts[character.name]
                for _, count in pairs(choices) do
                    count.current = 0
                end
            end
        )
    end
    local target = combatstate.round % 2 == 0 
        and M.targetmaxenergy(combatstate, character) 
        or M.targetminenergy(combatstate, character)
    local page
    local action
    if canattack then
        log.info(character.name .. " is attacking " .. target.name)
        action = "attack"
        local attackchoice = combatstate.choicecounts[character.name][action]
        if attackchoice.pagefun then
            page = attackchoice.pagefun(
                combatstate,
                nil,
                M.characterkey(combatstate, target)
            )
        else
            page = attack.attack(
                combatstate,
                nil,
                nil,
                M.characterkey(combatstate, target)
            )
        end
    end
    if not page and canshoot then
        log.info(character.name .. " is shooting " .. target.name)
        action = "shoot"
        page = attack.shoot(
            combatstate,
            nil,
            M.characterkey(combatstate, target)
        )
    end
    return M.decrementcount(page, character, action)
end


--[[
-- The average AI can use powers, and is smarter about target selection: it
-- targets the most vulnerable opponent, rather than the least. However, it
-- still targets the target with the highest energy for all damaging actions.
--
-- First, it cycles through its powers, using *only* those powers that do
-- damage. It targets the most vulnerable
-- character, with a bias towards targets with the most energy for damaging
-- powers.
--
-- Then it will attack the target with the highest energy as many times as it
-- has attacks left.
--
-- combatstate: The current state of combat
--
-- returns The combat page that the computer-controlled character should execute
-- ]]
function M.average(combatstate)
    local character = combatstate.characters[combatstate.charindex]
    log.info(character.name .. " is using the average AI.")
    local powercount = combatstate.choicecounts[character.name].power
    return (powercount and powercount.current > 0)
        and M.selectpower(combatstate, character)
        or M.dumb(combatstate)
end

--[[
-- Given the current state of combat, and a character, selects and returns a
-- page that executes the power the character will use.
--
-- combatstate - The current state of combat
-- character - The character who will use a power
--
-- return A page that executes the power selected for the character, and
-- decrements the power count
-- ]]
function M.selectpower(combatstate, character)
    if not combatstate.aistate then
        combatstate.aistate = {}
    end
    if not combatstate.aistate[character.name] then
        combatstate.aistate[character.name] = 0
    end
    local powernames = character.powers._keys
    local power = character.powers[
        powernames[combatstate.aistate[character.name]]
    ]
    for i, powername in ipairs(powernames) do
        local nextpower = character.powers[powername]
        assert(
            nextpower and nextpower.damage,
            "Power " .. powername .. " doesn't have a damage field."
        )
        if nextpower.damage > 0 and i > combatstate.aistate[character.name] then
            combatstate.aistate[character.name] = i % #powernames
            power = nextpower
        end
    end
    if not power then
        -- If this character doesn't have a damaging power, they just attack.
        return M.dumb(combatstate)
    end
    local targets = M.selectpowertargets(combatstate, character, power)
    local targetkeys = {}
    for _, target in ipairs(targets) do
        table.insert(targetkeys, M.characterkey(combatstate, target))
    end
    local powerpage = power.aiexecution(
        combatstate,
        nil,
        unpack(targetkeys)
    )
    return M.decrementcount(powerpage, character, "power")
end

-- [[
-- Given a combat page, returns a new page that
-- decrements the counter for the specified action before automatically
-- transitioning to and executing the specified combat page
--
--
-- combatpage - The page to execute
-- character - The character using the action
-- action - The name of the action whose count should be decremented
--
--
-- return - A page that decrements the character's power and transitions to
-- the power page
-- ]]
function M.decrementcount(combatpage, character, action)
    return pages.page(
        nil,
        function()
            return assert(
                combatpage,
                "decrementcount received `nil` for combatpage. If this is a "
                .. "power, you may have forgotten to import the new powers into "
                .. "data/powers.lua."
            )
        end,
        nil,
        function(combatstate)
            local choices = combatstate.choicecounts[character.name]
            choices[action].current = choices[action].current - 1
        end
    )
end

--[[
-- Selects a target for the specified character. Selecting a target is
-- deterministic. For powers that do heal damage, the target with the lowest
-- energy is targeted. For all other powers we take a weighted sum of the
-- difference between the user and target statistics for each statistic that
-- plays a role in the power. Statistics that are listed earlier in power.stats
-- are given a heavier weight. The character with the highest sum is then
-- targeted.
--
-- Powers that do damage also include the target's energy in the weighting, so
-- all else being equal, the AI will tend to target a character with higher
-- energy.
--
-- combatstate - The current state of combat
-- character - The character using the power
-- power - The power the character is using
--
-- return A list of between 0 and n targets, where n is the maximum number of
-- targets the specified power targets
-- ]]
function M.selectpowertargets(combatstate, character, power)
    local targets = {}
    local possibletargets = power.targets(combatstate)
    for _, target in pairs(possibletargets) do
        local targetsusceptibility = 0
        if power.damage > 0 then
            targetsusceptibility = target.statistics.energy
        end
        if power.damage < 0 then
            targetsusceptibility = 0 - target.statistics.energy
        else
            local stats = power.stats
            local numstats = #stats
            for i, stat in ipairs(stats) do
                local targetstat = target.statistics[stat]
                local characterstat = character.statistics[stat]
                assert(
                    characterstat,
                    "Character "
                        .. character.name
                        .. " does not have the statistic "
                        .. stat
                )
                local statweight = (characterstat - targetstat)
                    * (numstats - i - 1)
                targetsusceptibility = targetsusceptibility + statweight
            end
        end
        if #targets < power.numtargets then
            targets[targetsusceptibility] = target
        else
            local minsusceptibility = targetsusceptibility
            for othersusceptibility in pairs(targets) do
                if othersusceptibility < minsusceptibility then
                    minsusceptibility = othersusceptibility
                end
            end
            targets[targetsusceptibility] = target
            targets[minsusceptibility] = nil
        end
    end
    local targetlist = {}
    for _, target in pairs(targets) do
        table.insert(targetlist, target)
    end
    return targetlist
end

--[[
-- Returns the target with the highest energy
--
-- combatstate - The current state of combat
-- character - The character who is attacking
--
-- return The target with the highest energy
--]]
function M.targetmaxenergy(combatstate, character)
        local opponents = M.isally(combatstate, character)
            and combatstate.enemies
            or combatstate.allies
        local target = opponents[1]
        for _, opponent in ipairs(opponents) do
            if target.statistics.energy < opponent.statistics.energy then
                target = opponent
            end
        end
        return target
end


--[[
-- Returns the target with the lowest energy
--
-- combatstate - The current state of combat
-- character - The character who is attacking
--
-- return The target with the lowest non-zero energy, unless
-- all targets have zero energy, in which case returns the first
-- target
--]]
function M.targetminenergy(combatstate, character)
        local opponents = combatstateops.activeopponents(combatstate)
        local target = opponents[1]
        local targetenergy = target.statistics.energy
        for _, opponent in ipairs(opponents) do
            local bettertarget = targetenergy > opponent.statistics.energy
            if bettertarget then
                target = opponent
                targetenergy = opponent.statistics.energy
            end
        end
        return target
end


--[[
-- Marks this character as having completed their last action for the round.
--
-- We do that by setting this character's action counts for every action
-- for the round to 0.
--
-- combatstate: The current state of combat
-- charactername: The name of the character whose round should end
-- ]]
function M.lastaction(combatstate, charactername)
    local characterchoices = combatstate.choicecounts[charactername]
    for choice, _ in pairs(characterchoices) do
        characterchoices[choice] = 0
    end
end

--[[
-- Finds the Love KeyConstant that selects the specified character
--
-- combatstate - The current state of combat
-- character - The character to find the KeyConstant for
--
-- return The KeyConstant that selects the specified character
-- ]]
function M.characterkey(combatstate, character)
    if LOVE then
        for index, ally in pairs(combatstate.allies) do
            if ally == character then
                return "f" .. index
            end
        end
        for index, enemy in pairs(combatstate.enemies) do
            if enemy == character then
                return tostring(index)
            end
        end
    else
        for index, allyenemy in pairs(combatstate.characters) do
            if character == allyenemy then
                return tostring(index)
            end
        end
    end
    local errorstring = "Unknown character " .. character.name
    log.error(errorstring)
    assert(false, errorstring)
end

--[[
-- Returns true if the specified character is an ally (i.e. working with the
-- player), and false otherwise
--
-- combatstate - The current state of comat
-- character - The character who may or may not be an ally
--
-- returns true if the character is an ally, false otherwise
--]]
function M.isally(combatstate, character)
    for _, ally in pairs(combatstate.allies) do
        if ally.name == character.name then
            return true
        end
    end
    return false
end

return M
