--[[
-- Copyright 2018 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
-- ]]
--
-- [[
-- Contains a collection of utility functions for working with statuses.
-- ]]

local M = {}
local combatstateops = require 'combat/combatstate'
local F = require 'f-strings/F'
local log = require 'log'
local scarlettables = require 'scarlettables'

-- [[
-- Increases the specified character's attrition by the specified
-- amount.
--
-- combatstate: The current state of combat
-- character: The character whose attrition should be increased
-- increment: The amount to increase attrition by, defaults to 1
-- ]]
function M.increaseattrition(combatstate, character, increment)
    increment = increment or 1
    local attrition = character.attrition
    character.attrition = attrition + increment
end

-- [[
-- Returns the table containing the specified status (based on name), and how many more
-- rounds
-- the character is inflicted with that status. If nil is returned, then the
-- character is not afflicted with that status
--
-- character: The character to check
-- status: Either the status to look for, or the name of the status to look for
--
-- return Two values: A table the contains the status (`status`)
-- and how many more rounds the character is inflicted with the specified
-- status (`duration), and its index in the character's list of statuses.
-- Returns nil if the character is not inflicted with the specified status
-- ]]
function M.statusduration(character, status)
    local statusname = type(status) == "string" and status or status.name
    for i, statusduration in pairs(character.statuses) do
        if statusduration.status.name == statusname then
            return statusduration, i
        end
    end
end

--[[
-- Adds the specified status to the specified character's status list for the
-- specified duration, stacking them if the character already has that 
-- status.
--
-- character: The character being inflicted with a status
-- status: The status being inflicted
-- duration: How long the character is afflicted with the status
-- maybeuser: The character who inflicted the status. Whether or not
    a status needs this depends on the status
-- maybebuff: A number representing the "potency" of the buff. Whether or
--  not a status needs this depends on the status
-- combatstate: The current state of combat. Whether a status needs this
-- or not depends on the status
--]]
function M.addstatus(
    character,
    status,
    duration,
    maybeuser,
    maybebuff,
    combatstate
)
    local buff
    local user
    -- In this case, maybebuff is the combatstate
    if not combatstate then
        combatstate = maybebuff or maybeuser
        user = maybebuff and maybeuser or nil
        buff = 0
    else
        user = maybeuser
        buff = maybebuff
    end
    assert(
        duration,
        "A status needs a duration! : " .. scarlettables.string(status)
    )
    local statusduration = M.statusduration(character, status.name)
    if statusduration and statusduration.status.notstackable then
        return
    elseif statusduration then
        if statusduration.status.stack then
            statusduration.status.stack(character, combatstate, user, buff)
        end
        statusduration.duration = statusduration.duration + duration
    else
        status.apply(character, combatstate, user, buff)
        assert(
            buff ~= combatstate,
            "Trying to add a buff number that's actually the combatstate!"
        )
        table.insert(
            character.statuses, 
            {status=status, duration=duration, buff=buff}
        )
        table.insert(
            combatstate.combatlog,
            F"{character.combatname} gains the {status.name} status!"
        )
    end
end

-- [[
-- Removes the specified status from the character's status list. If the
-- character is not afflicted with the specified status, then this is a
-- noop.
--
-- character: The character to remove a status from
-- status: The status (or statusduration) to remove
-- ]]
function M.removestatus(character, status, combatstate)
    status = status.status or status
    local statusduration, i = M.statusduration(character, status)
    if i then
        if statusduration.status.remove then
            statusduration.status.remove(character, combatstate)
        end
        table.remove(character.statuses, i)
        table.insert(
            combatstate.combatlog,
            F"{status.name} fades from {character.combatname}!"
        )
    end
end

--[[
-- Decrements the specified status on the specified character by
-- the specified amount, removing it if the duration drops to 0
-- or lower.
-- status: The status to reduce
-- character: The character who has been inflicted with this status
-- decrement: The amount to reduce the status duration by
-- combatstate: The current state of combat
--]]
function M.decrementstatus(status, character, decrement, combatstate)
    local statusduration = assert(M.statusduration(character, status))
    statusduration.duration = statusduration.duration - decrement
    if statusduration.duration <= 0 then
        M.removestatus(character, status, combatstate)
    end
end

--[[
-- Decrements all statuses on the specified character by the 
-- specified amount, removing each status if its duration drops
-- to 0 or lower.
-- 
-- character: The character whose statuses should be decremented.
-- combatstate: The current state of combat
-- decrement: The amount to decrement each status' duration by
--]]
function M.decrementstatuses(character, combatstate, decrement)
    log.info("Decrementing statuses for " .. character.name)
    -- We make a shallow copy, because decrementstatuses modifies
    -- character.statuses.
    local statuses = scarlettables.shallowcopy(character.statuses)
    for _, statusduration in ipairs(statuses) do
        local status = statusduration.status
        M.decrementstatus(status, character, decrement, combatstate)
    end
end

-- [[
-- Increases the max and current count for the specified choice by the
-- specified increment.
--
-- choices: An ordered table of choices to modify. Table maps choice names to
--  a table with three keys:
--      max - The maximum number of times this choice can be used
--      current - The remaining number of times this choice can be used
--      page - The combat page to execute when this choice is selected
-- choicename: The name of the choice whose count should be incremented
-- increment: The amount to increase the choice count by, defaults to 1
-- actionpage: The page to associate with this choice, this is only necessary
--  if a new choice is being added
-- ]]
function M.increasechoice(choices, choicename, increment, actionpage)
    increment = increment or 1
    local choice = choices[choicename]
    if not choice then
        assert(
            actionpage,
            "Attempting to add a new choice "
                .. choicename
                .. " but no actionpage was provided."
        )
        choices[choicename] = {
            max=increment,
            current=increment,
            page=actionpage
        }
        table.insert(choices._keys, choicename)
    else
        choice.max = choice.max + increment
        choice.current = choice.current + increment
    end
end

--[[
-- Decreases the max and current count for the specified choice by the
-- specified amount.
--
-- Note that the choice must already exist before you can decrement it.
--
-- choices: The choices to modify
-- choicename: The name of the choice to decrement
-- decrement: The amount to decrement the current and max count by
--]]
function M.decreasechoice(choices, choicename, decrement)
    M.increasechoice(choices, choicename, -decrement)
end


--[[
-- Given a character, returns the status with the best buff, or
-- nil if there is none (i.e. all the statuses are debuffs).
--
-- character: The character whose best buff is desired
--
-- returns The best buff the character has and its duration, or nil if they
-- don't have any buffs with at least a 0 rating
--]]
function M.bestbuff(character)
    local buffvalue = 0
    local bestbuff
    for _, statusduration in ipairs(character.statuses) do
        local statusbuff = statusduration.status.buff
        if type(statusbuff) == "function" then
            statusbuff = statusbuff(character)
        end
        if statusbuff >= buffvalue then
            bestbuff = statusduration
            buffvalue = statusbuff
        end
    end
    if bestbuff then
      return bestbuff.status, bestbuff.duration
    end
end

-- [[
-- Returns the first opponent who isn't afflicted with the
-- provided status, or nil if there is none.
--
-- opponents: The list of characters to check
-- status: The status to look for
--
-- return The first opponent who does not have the specified status,
-- or nil if every character is affected
function M.unaffectedopponent(opponents, status)
    for _, opponent in ipairs(opponents) do
        if not M.statusduration(opponent, status) then
            return opponent
        end
    end
    return nil
end

--[[
-- Returns the first opponent who is afflicted with the
-- provided status, or nil if there is none
--
-- opponents: The list of characters to check
-- status: The status to look for
--
-- return THe first opponent who is affected by the specified status,
-- or nil if nobody is affected
--]]
function M.affectedopponent(opponents, status)
    for _, opponent in ipairs(opponents) do
        if M.statusduration(opponent, status) then
            return opponent
        end
    end
    return nil
end

--[[
-- Executes all of the specified character's statuses' start_of_turn operations
--
-- They are executed in the order they were inflicted.
--
-- character: The character whose statuses' start of turn operation should be 
--  executed.
-- combatstate: The current state of combat
--]]
function M.startturn(character, combatstate)
    -- Note that some statuses may remove other statuses automatically,
    -- and we don't want the list we're iterating over to change under
    -- our feet. This does mean that all statuses trigger their startturn,
    -- even if another status removes it.
    local statuses = scarlettables.shallowcopy(character.statuses)
    for _, statusduration in ipairs(statuses) do
        local status = statusduration.status
        -- A previous status' startturn may have removed this status.
        if M.statusduration(character, status) and status.startturn then
            status.startturn(character, combatstate)
        end
    end
end

--[[
-- Executes all of the specified character's statuses' endturn operations
--
-- They are executed in the order they were inflicted.
--
-- character: The character whose statuses' start of turn operation should be 
-- executed.
-- combatstate: The current state of combat
--]]
function M.endturn(character, combatstate)
    local statuses = scarlettables.copy(character.statuses)
    for _, statusduration in ipairs(statuses) do
        local status = statusduration.status
        if status.endturn then
            status.endturn(character, combatstate)
        end
    end
end

return M
