--[[
-- Copyright 2021 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]

--[[
-- Implements Bright's powers. Bright is a lightning fast glass cannon.
She hits really hard and has a sky-high speed (and thus very high defense),
but fairly low energy and with the exception of Brightwings, none of her
skills are defensive in nature. In fact some of them even reduce her defense.
--]]


local attacks = require 'combat/attack'
local combatoperations = require 'combat/combatstate'
local F = require 'f-strings/F'
local log = require 'log'
local pages = require 'pages'
local powers = require 'combat/powers'
local scarletmoonpowers = require 'data/scarletmoonpowers'
local scarlettables = require 'scarlettables'
local statuses = require 'data/statuses'
local statusops = require 'combat/statuses'
local targeting = require 'combat/target'
local powerutils = require 'data/powerutils'

local M = {}

---- Brightsaber ----
--[[
-- Increases Bright's attack count by 1.
--]]
M.brightsaber =  pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local character = combatoperations.activecharacter(combatstate)
        local combatname = character.combatname
        local power = attacks.statistic(character, "power")
        local buff = math.max(1, power)
        table.insert(combatstate.combatlog, F"A Glowing saber forms in {combatname}'s hand!")
        statusops.addstatus(
            character,
            statuses.attack,
            power,
            character,
            buff,
            combatstate
        )
    end
)
M.brightsaber.cost = 1
M.brightsaber.damage = 0
M.brightsaber.targets = combatoperations.activeallies
M.brightsaber.numtargets = 0
M.brightsaber.stats = {"power"}
function M.brightsaber.description(character)
    return F"Bright creates a saber made of Glow for herself, increasing"
        .. F" her Attack count, and attrition by 1."
end


---- Brightwings ----
--[[
-- Grants Bright the Fly status, which increases her defense
-- by power/2 + speed/2, but doubles the damage she takes from Twister, and how far she gets
-- sent flying by Airburst.
--]]
M.brightwings =  pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local character = combatoperations.activecharacter(combatstate)
        local combatname = character.combatname
        local power = attacks.statistic(character, "power")
        local speed = attacks.statistic(character, "speed")
        local buff = math.max(1, math.floor(speed/2))
        table.insert(combatstate.combatlog, F"Glowing wings emerge from {combatname}'s back!")
        statusops.addstatus(
            character,
            statuses.fly,
            power,
            character,
            buff,
            combatstate
        )
    end
)
M.brightwings.cost = 1
M.brightwings.damage = 0
M.brightwings.targets = combatoperations.activeallies
M.brightwings.numtargets = 0
M.brightwings.stats = {"power", "speed"}
function M.brightwings.description(character)
    local speed = attacks.statistic(character, "speed")
    local buff = math.floor(speed/2)
    return F"Bright creates wings of Glow and flies with them,"
        .. F" increasing her Attack Defense by *{buff}*, but "
        .. F" doubling the impact of wind-based powers like Twister."
end

local function resolvetarget(targetkey, combatstate)
    if LOVE then
        if string.sub(targetkey, 1, 1) == "f" then
            targetkey = string.sub(targetkey, 2)
            return combatstate.allies[tonumber(targetkey)]
        else
            return combatstate.enemies[tonumber(targetkey)]
        end
    else
        return combatstate.characters[tonumber(targetkey)]
    end
end

---- Flashing Cleave
--[[
-- Flashing Cleave allows Bright to destroy a target's shield.
--]]
function M.invokeflashingcleave(combatstate, previouspage, targetkey)
    local user = combatoperations.activecharacter(combatstate)
    local target = resolvetarget(targetkey, combatstate)
    local targetable = target
        and not combatoperations.isdefeated(combatstate, target.name)
    return pages.page(
        nil,
        function() if not targetable then return previouspage end end,
        nil,
        function(combatstate1)
            if not targetable then
                return
            end
            -- Do exactly as much damage as remains on the target's
            -- shield, effectively destroying it.
            attacks.shield(target, attacks.shield(target))
            table.insert(
                combatstate1.combatlog,
                user.combatname .. " slices through " .. target.combatname 
                    .. "'s shield with a flashing cleave!"
            )
        end
    )
end

M.flashingcleave = targeting.target(M.invokeflashingcleave, 1)
M.flashingcleave.cost = 5
M.flashingcleave.damage = 0
M.flashingcleave.targets = combatoperations.activeenemies
M.flashingcleave.numtargets = 1
M.flashingcleave.stats = {"power"}
function M.flashingcleave.description(character)
    return F"Bright destroys a target's shield with a flashing cleave."
end


--- Blinding Charge
--[[
-- Bright launches a reckless, lightning-fast charge at a target. She
-- does damage equal to attack + defense/2, and reduces her defense by
-- half until the start of her next turn. If she is Away, she cures 
-- away, deals attack + defense damage, and reduces her defense to 0
-- until the start of her next turn.
--]]

function M.invokeblindingcharge(combatstate, previouspage, targetkey)
    local target = resolvetarget(targetkey, combatstate)
    log.info("Resolved target key " .. targetkey .. " to " 
        .. (target and target.name or "nil")
    )
    local user = combatoperations.activecharacter(combatstate)
    local targetable = target
        and not combatoperations.isdefeated(combatstate, target.name)
    return pages.page(
        nil,
        function() if not targetable then return previouspage end end,
        nil,
        function(combatstate1)
            if not targetable then
                log.info(targetkey .. " not targetable. Doing nothing.")
                return
            end
            local defense = attacks.defensestat(user)
            local attackbonus = statusops.statusduration(user, statuses.away)
                and defense
                or math.floor(defense / 2)
            local damage = attacks.attackstat(user) + attackbonus
            local function chargedamage(attacker, defender, combatstate2)
                return attacks.defaultdamage(
                    attacker,
                    defender,
                    combatstate2,
                    function() return damage end
                )
            end
            table.insert(
                combatstate1.combatlog,
                F"{user.combatname} charges recklessly at {target.combatname}"
                    .. " in a burst of bright light!"
            )
            attacks.damage(combatstate1, chargedamage, {target},  user)
            statusops.removestatus(user, statuses.away, combatstate1)
            statusops.addstatus(
                user,
                statuses.vulnerable,
                2,
                user,
                attackbonus,
                combatstate1
            )
        end
    )
end

M.blindingcharge = targeting.target(M.invokeblindingcharge)
M.blindingcharge.cost = 10
M.blindingcharge.damage = powers.DAMAGES
M.blindingcharge.targets = combatoperations.activeenemies
M.blindingcharge.numtargets = 1
M.blindingcharge.stats = {"strength"}
function M.blindingcharge.description(character)
    local defense = attacks.defensestat(character)
    local bonus = math.floor(defense / 2)
    local attack = attacks.attackstat(character)
    local damage = attack + bonus
    local awaydamage = defense + damage
    return F"Bright charges recklessly at a target, dealing *{damage}* damage,"
        .. F" and reducing her defense by *{bonus}* for 1 turn. If Bright is "
        .. F"Away, she deals *{awaydamage}* instead, her defense drops to *0* and is cured of Away."
end

return M
