--[[
-- Copyright 2018 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]

--[[
-- Defines an ordered table containing-all the options (like AI intelligence,
-- attrition level, that sort of thing) available in the game.
--
-- Furthermore, don't forget to add the name of each option to the _keys list.
-- The engine relies on the _keys field to provide a consistent ordering of the
-- options. If the name isn't included in _keys, then it won't be displayed,
-- or used in any way.
--
-- The engine automatically sets `gamestate.options[optionname]` to the name
-- of the currently selected value. So if your option relies solely on
-- `gamestate.options[optionname]` (i.e. AI), then your apply doesn't need to
-- do anything
--
-- See utilities/options.md for details on the structure of options.
--]]
--
local function donothing() end

-- [[
-- Builds a list of numbers starting at `from` and going to `to` with `step`
-- incrments.
--
-- from: The starting number of the list
-- to: The ending number of the list
-- step: The increments between entries in the list, defaults to 1
-- ]]
local function range(from, to, step)
    step = step or 1
    local list = {}
    local current = from
    while (to - current >= step) do
        table.insert(list, current)
        current = current + step
    end
    return list
end

--[[
-- Given the name of a font and a number range, returns an option that allows
-- the player to select any number in the range as the font size.
--
-- fontname: The name of the type of font (i.e. title, storybody, etc) that we
-- should build sizes for
-- fontsizes: A list of available font sizes
--
-- return An option with a choice for every number in the range
--]]
local function buildfont(startsize, fontsizes)
    local sizeoptions = {}
    local startingindex = 1
    for i, size in ipairs(fontsizes) do
        if size == startsize then
            startingindex = i
        end
        table.insert(
            sizeoptions,
            {
                name=tostring(size),
                apply=donothing,
                remove=donothing
            }
        )
    end
    sizeoptions.startindex = startingindex
    return sizeoptions
end

local fontsizes = range(8, 33)

local options = {
    _keys={
        "ai",
        "attrition",
        "spankableplayer",
        "combat",
        "bumstatus"
    },
    ai={
        {
            name="smart",
            apply=donothing,
            remove=donothing
        },
        {
            name="average",
            apply=donothing,
            remove=donothing
        },
        {
            name="dumb",
            apply=donothing,
            remove=donothing
        }
    },
    attrition={
        {
            name="lose 3% energy each round",
            apply=function(gamestate)
                function gamestate.attrition(character)
                    return math.ceil(character.statistics.maxenergy * .03)
                end
            end,
            remove=donothing
        },
        {
            name="lose 2% energy each round",
            apply=function(gamestate)
                function gamestate.attrition(character)
                    return math.ceil(character.statistics.maxenergy * .02)
                end
            end,
            remove=donothing
        },
        {
            name="lose 0% energy each round",
            apply=function(gamestate)
                function gamestate.attrition()
                    return 0
                end
            end,
            remove=donothing
        }
    },
    spankableplayer={
        {
            name="yes",
            apply=donothing,
            remove=donothing
        },
        {
            name="no",
            apply=donothing,
            remove=donothing
        },
    },
    combat={
        {
            name="yes",
            apply=donothing,
            remove=donothing
        },
        {
            name="no",
            apply=donothing,
            remove=donothing
        }
    },
    bumstatus={
        {
            name="emoji",
            apply=donothing,
            remove=donothing
        },
        {
            name="text",
            apply=donothing,
            remove=donothing
        }
    },
    titlefont=LOVE and buildfont(32, fontsizes) or nil,
    bodyfont=LOVE and buildfont(20, fontsizes) or nil,
    combatfont=LOVE and buildfont(22, fontsizes) or nil,
    buttonfont=LOVE and buildfont(20, fontsizes) or nil,
}

if LOVE then
    table.insert(options._keys, "titlefont")
    table.insert(options._keys, "bodyfont")
    table.insert(options._keys, "combatfont")
    table.insert(options._keys, "buttonfont")
end

return options
