--[[
-- Copyright 2018 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]
--
--[[
-- This module defines the smart actions for Prometheus and Poseidon in Scarlet
-- Moon.
--
-- Prometheus and Poseidon's main gimmick is a ladder of four statuses that
-- help them and hurt their enemies:
--
--  1. Frozen - The worst "cold" status. Anybody other than Poseidon who is
--      frozen has their attack count reduced to 0, and their speed cut in half.
--      Poseidon has his strength increased by Prometheus' power / 2 and
--      damage reduction of the same amount.
--      A character needs to be wet before they can be frozen. Prometheus
--      can inflict this status on a Wet character when using one of her self
--      buffs. Prometheus' aggressive powers replace Frozen with Wet.
--  2. Wet - The character is covered in water. Poseidon's attacks do extra
--      damage to them. Poseidon needs to be Wet for some of his deadlier
--      powers to trigger. Prometheus' aggressive powers remove Wet.
--  3. Warm - Inflicted whenever Prometheus uses an aggressive power on an
--      enemy. Warm characters take extra damage from Prometheus. Prometheus
--      needs to be Warm for some of her more powerful powers to trigger.
--      Poseidon's aggressive powers replace Warm with Wet.
--  4. Aflame - The character is on fire. They take damage with each action,
--      and Prometheus' powers do double damage against them. Some of
--      Prometheus' more powerful powers can inflict Aflame on a Warm character.
--      Poseidon's aggressive powers can replace Aflame with Warm.
--      If Prometheus is Aflame, she does double damage, and can use her most
--      dangerous powers.
--]]

local M = {}

local ai = require 'combat/ai'
local attacks = require 'combat/attack'
local combatstateops = require 'combat/combatstate'
local statuses = require 'data/statuses'
local statusops = require 'combat/statuses'
local log = require 'log'
local powers = require 'data/powers'
local powerutils = require 'data/powerutils'

--[[
-- Prometheus' AI for episode 3 day 1.
--
--Prometheus is a major damage dealer. She also buffs Poseidon and herself, and
--can inflict statuses that increase the damage enemies take from her powers.
--Her pattern is as follows:
--  1. If her buff isn't active, use twin serpents to get her buff. If a
--  character is Wet, target them (prioritizing Poseidon, otherwise opponents),
--  and make them Frozen.
--  2. Otherwise, use her fireball power on a Warm hero, prioritizing the
--  character with middle energy levels (so neither the lowest, nor the
--  highest).
--  3. Once all her powers are used up, attack the character with the highest
--  energy.
--
--  If Prometheus is Wet, she can't use either of her powers.
--]]
function M.prometheusep3d1(combatstate)
    log.info("Using Prometheus ep3d1 AI.")
    local character = combatstateops.activecharacter(combatstate)
    local canusepowers =
        combatstate.choicecounts[character.name].power.current > 0
        and not statusops.statusduration(character, statuses.wet)
    if canusepowers then
        log.info("Using powers.")
        return statusops.statusduration(character, statuses.twinserpents)
            and ai.decrementcount(powers.fireball, character, "power")
            or ai.decrementcount(powers.twinserpents, character, "power")
    end
    return ai.dumb(combatstate)
end

function M.prometheusep3d2(combatstate)
    log.info("Using Prometheus' ep3d2 AI.")
    local character = combatstateops.activecharacter(combatstate)
    local canusepowers =
        combatstate.choicecounts[character.name].power.current > 0
        and not statusops.statusduration(character, statuses.wet)
    local hassummon = powerutils.issummon(character)
        or powerutils.findsummon(combatstate, character)
    if not hassummon then
        return ai.decrementcount(powers.flamey, character, "power")
    else
        return M.prometheusep3d1(combatstate)
    end
end

--[[
-- Poseidon's AI for episode 3 day 1.
--
-- Poseidon has two modes: Wet and Frozen. While Wet, Poseidon primarily relies
-- on firing jets of water at enemies, and enhancing his (low) speed with a
-- water slide. Though his water slide does require him to cure a character of
-- Wet if there is one other than himself. When he is Frozen he becomes a very
-- tough bruiser, with high damage reduction, high strength, a frozen punch,
-- and a retaliatory power.
--
-- However, if Poseidon becomes warm then his powers are useless.
--
-- His pattern is as follows:
--
-- 1. If his buff isn't active, use water slide to get his buff. If a character
-- is Wet, target them (prioritizing allies).
--
-- 2. If Poseidon is Wet, use his water hammer power to attack the hero with
-- health closest to the median. Otherwise, use hydroslide to get wet.
--
-- 3. If Poseidon is Frozen, and not buffed, use his buff.
--
-- 4. Otherwise attack the hero with the median health with his frozen punch
-- power, and the highest energy character with his regular attack.
--
--]]
function M.poseidonep3d1(combatstate)
    log.info("Using Poseidon episode 3 day 1 AI.")
    local character = combatstateops.activecharacter(combatstate)
    local canusepowers =
        combatstate.choicecounts[character.name].power.current > 0
        and not statusops.statusduration(character, statuses.warm)
    if canusepowers then
        log.info("Using powers.")
        if statusops.statusduration(character, statuses.frozen) then
            log.info(attacks.retaliation(character) <= 0 and "Using chionskin" or "Using chionfist.")
            return attacks.retaliation(character) <= 0
                and ai.decrementcount(powers.chionskin, character, "power")
                or ai.decrementcount(powers.chionfist, character, "power")
        else
            local hydroslide = statusops.statusduration(
                character,
                statuses.hydroslide
            )
            local wet = statusops.statusduration(character, statuses.wet)
            return (hydroslide and wet)
                and ai.decrementcount(powers.hydrohammer, character, "power")
                or ai.decrementcount(powers.hydroslide, character, "power")
        end
    end
    return ai.dumb(combatstate)

end


function M.poseidonep3d2(combatstate)
    local character = combatstateops.activecharacter(combatstate)
    log.info("Using Poseidon's ep3d2 AI for " .. character.name)
    local canusepowers =
        combatstate.choicecounts[character.name].power.current > 0
        and not statusops.statusduration(character, statuses.warm)
    local hassummon = powerutils.issummon(character)
        or powerutils.findsummon(combatstate, character)
    if not hassummon and canusepowers then
        return ai.decrementcount(powers.hydrogator, character, "power")
    else
        return M.poseidonep3d1(combatstate)
    end
end

return M
