--[[
-- Copyright 2017 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]
--
--[[
-- This module defines the smart actions for computer-controlled characters
-- in Scarlet Moon.
--]]

local M = {}

local ai = require 'combat/ai'
local attacks = require 'combat/attack'
local combatstateops = require 'combat/combatstate'
local scarletmoonpowers = require 'data/scarletmoonpowers'
local statuses = require 'data/statuses'
local statusops = require 'combat/statuses'
local pages = require 'pages'
local powers = require 'data/powers'
local powerutils = require 'data/powerutils'
local log = require 'log'

local prometheusposeidon = require 'data/prometheusposeidon'

for name, func in pairs(prometheusposeidon) do
    M[name] = func
end

-------------------- Buzzsaw -----------------------------------------------
--[[
-- Buzzsaw's starting AI. Buzzsaw has the following pattern:
--
-- 1. If Buzzsaw is not already buffed with chainwhip, then she buffs herself
-- with chainwhip.
-- 2. Otherwise, she uses her Buzzsaw power on the enemy with the highest
-- energy, and she attacks the enemy with the highest energy.
--
-- return The combat page to execute
--]]
function M.buzzsaw(combatstate)
    local character = combatstateops.activecharacter(combatstate)
    log.info(character.name .. " is using Buzzsaw's AI.")
    local powercount = combatstate.choicecounts[character.name].power
    if powercount.current > 0 then
        if statusops.statusduration(character, statuses.chainwhip) then
            local target = ai.targetmaxenergy(combatstate, character)
            log.info("Targeting " .. target.name .. " with buzzsaw.")
            local targetkey = ai.characterkey(combatstate, target)
            return ai.decrementcount(
                powers.buzzsaw.aiexecution(combatstate, nil, targetkey),
                character,
                "power"
            )
        else
            log.info("Targeting self with chainwhip.")
            return ai.decrementcount(
                powers.chainwhip.aiexecution(combatstate),
                character,
                "power"
            )
        end
    end
    log.info("All powers used this round. Attacking enemy with highest energy.")
    return ai.dumb(combatstate)
end

--[[
-- Buzzsaw's AI for the final battle of episode 1. Buzzsaw has the following
-- patterns:
--
-- 1. If Buzzsaw is not already buffed with SoaringLeap, then she buffs herself
-- with soaring leap.
-- 2. If Buzzsaw just buffed herself with chainwhip, then she binds an enemy
-- with chainpull.
-- 3. Otherwise, she behaves just like M.buzzsaw.
--
-- combatstate: The current state of combat
--
-- return The combat page for Buzzsaw to execute
--]]
function M.buzzsaw_ep1d2(combatstate)
    local character = combatstateops.activecharacter(combatstate)
    log.info(character.name .. " is using Buzzsaw's ep1d2 AI.")
    local powercount = combatstate.choicecounts[character.name].power
    if powercount.current > 0 then
        if not statusops.statusduration(character, statuses.leap) then
            log.info("Buffing self with soaring leap.")
            return ai.decrementcount(
                powers.soaringleap.aiexecution(combatstate),
                character,
                "power"
            )
        end
        local chainwhipstatus = statusops.statusduration(
            character,
            statuses.chainwhip
        )
        local justusedchainwhip = chainwhipstatus
            and chainwhipstatus.duration >= 2*character.statistics.power - 1
        if justusedchainwhip then
            log.info("Debuffing enemy with chain pull.")
            return ai.decrementcount(
                powers.chainpull.aiexecution(combatstate),
                character,
                "power"
            )
        end
    end
    return M.buzzsaw(combatstate)
end

---------------------------- Succubus ------------------------------------

--[[
-- Given a list of characters and a status, returns the first character with
-- the status, or nil if there are none.
--
-- targets: The list of characters which may or may not have a character
--  with the specified status
--  checkedstatus: The status to look for
--
--  return The first character found with the status, or nil if there are none
--]]
local function status(targets, checkedstatus)
    for _, target in ipairs(targets) do
        if statusops.statusduration(target, checkedstatus) then
            return target
        end
    end
end

--[[
-- Performs an attack that is identical to a regular attack against the
-- specified target, except the description is replaced that of the attacker
-- spanking their opponent.
--
-- combatstate: The current state of combat
-- attacker: The attacker
-- target: The target to be spanked
--
-- return A combat page that executes an attack with a succubus-specific
-- description
--]]
local function spankattack(combatstate, attacker, target)
    local choices = combatstate.choicecounts[attacker.name]
    if choices.attack.current <= 0 then
        return pages.page(
            nil,
            nil,
            nil,
            function()
                for _, choice in pairs(choices) do
                    choice.current = 0
                end
            end
        )
    end
    local function damage(attacker1, target1, combatstate1)
        local damage, retaliation = attacks.defaultdamage(
            attacker1,
            target1,
            combatstate1
        )
        local description = attacker1.combatname .. " yanks on "
            .. target1.combatname .. "'s hair and slaps "
            .. target1.combatname .. "'s raised, wiggling bottom "
            .. (damage > 0 and (damage .. " times! ") or "many times! ")
            .. target1.combatname .. " kicks and flails helplessly!"
        if retaliation and retaliation > 0 then 
            description = description 
                .. F"{target1.combatname} retaliates for {retaliation} damage!"
        end
        return damage, retaliation, {description}
    end
    return pages.page(
        nil,
        nil,
        nil,
        function(combatstate1)
           attacks.damage(combatstate1, damage, {target}, attacker)
        end
    )
end

--[[
-- An AI for Succubus.
--]]
function M.succubus(combatstate)
    local character = combatstateops.activecharacter(combatstate)
    local targets = combatstateops.opponents(combatstate, character)
    local spankedtarget = status(targets, statuses.spanked) 
        or status(targets, statuses.succuspanked)
    local energy = character.statistics.energy
    local multipleofthree = combatstate.round % 3 == 0
    if combatstate.choicecounts[character.name].power.current > 0 then
        local otk = status(targets, statuses.otk)
        local succuspank = energy > powers.succuspank.cost 
            and otk 
            and not spankedtarget 
            and not multipleofthree
        local overmyknee = energy > powers.overmyknee.cost 
            and not otk 
            and not multipleofthree
            and not statusops.statusduration(character, statuses.spanked)
        if succuspank then
            log.info("Administering a succuspanking.")
            usedsuccuspank = true
            return ai.decrementcount(powers.succuspank, character, "power")
        elseif overmyknee then
            log.info("Putting a character over Succubus' knee.")
            usedsuccuspank = true
            return ai.decrementcount(
                powers.overmyknee,
                character,
                "power"
            )
        elseif energy > powers.cropwhip.cost then
            log.info("Administering a whipping with a riding crop.")
            usedsuccuspank = false
            return ai.decrementcount(powers.cropwhip, character, "power")
        end
    end
    local otktarget = status(targets, statuses.otk)
    return otktarget
        and ai.decrementcount(
            spankattack(combatstate, character, otktarget),
            character,
            "attack"
        )
        or ai.dumb(combatstate)
end

local leopardcount = 0
--[[
-- An AI for Leopard.
--
-- Leopard is a very aggressive character, and savvy opponent, who
-- won't hestitate to turn an opponent's strength against them.
--
-- Her basic pattern is hitting someone with distortion, following
-- up with a pair of roundhouse kicks, alternating between kicking
-- and throwing for four turns, and then finally going back to 
-- distortion.
--]]
function M.leopardep5d2(combatstate)
    local character = combatstateops.activecharacter(combatstate)
    log.info(character.name .. " is using Leopard's AI.")
    local powercount = combatstate.choicecounts[character.name].power
    if powercount.current > 0 then
        if combatstate.round == 1 then
            return ai.decrementcount(powers.judo, character, "power")
        end
        leopardcount = leopardcount + 1
        if leopardcount > 4 then
            leopardcount = 1
        end
        if leopardcount == 1 then
            return ai.decrementcount(powers.distortion, character, "power")
        elseif leopardcount == 2 then
            return ai.decrementcount(powers.roundhousekick, character, "power")
        else
            return ai.decrementcount(powers.throw, character, "power")
        end
    end
    return ai.dumb(combatstate)
end

--[[
-- An AI for Bright. Bright is a very aggressive character, who tries to
-- take you down as quickly as possible. She starts off by buffing herself
-- with Brightwings and Brightsaber. Then she  will use Flashing Cleave
-- whenever Scarlet Moon has a shield, but otherwise uses Blinding Charge.
-- If he's Away, she prioritizes Blinding Charge.
--]]
function M.brightep6d1(combatstate)
    local character = combatstateops.activecharacter(combatstate)
    log.info(character.name .. " is using Bright's AI.")
    local powercount = combatstate.choicecounts[character.name].power
    if powercount.current > 0 then
        if not statusops.statusduration(character, statuses.fly) then
            return ai.decrementcount(powers.brightwings, character, "power")
        end
        if not statusops.statusduration(character, statuses.attack) then
            return ai.decrementcount(powers.brightsaber, character, "power")
        end
        local shieldedopponents = {}
        local opponents = combatstateops.activeopponents(combatstate)
        for _, character in pairs(opponents) do
            if attacks.shield(character) > 0 then
                log.info("opponent found with a shield greater than zero.")
                table.insert(shieldedopponents, character)
            end
        end
        local charge = statusops.statusduration(character, statuses.away) 
            or not next(shieldedopponents) 
        if charge then
            local target = ai.targetmaxenergy(combatstate, character)
            log.info("Targeting " .. target.name .. " with blinding charge.")
            local targetkey = ai.characterkey(combatstate, target)
            log.info("Using target key " .. targetkey .. " with blinding charge.")
            local powerpage = powers.invokeblindingcharge(
                combatstate,
                nil,
                targetkey
            )
            return ai.decrementcount(powerpage, character, "power")
        end
        if next(shieldedopponents) then
            local maxshield = 0
            local target
            for _, opponent in pairs(shieldedopponents) do
                local shield = attacks.shield(opponent)
                if shield > maxshield then
                    target = opponent
                    maxshield = shield
                end
            end
            log.info("Targeting " .. target.name .. " with flashing cleave.")
            local targetkey = ai.characterkey(combatstate, target)
            local powerpage = powers.invokeflashingcleave(
                combatstate,
                nil,
                targetkey
            )
            return ai.decrementcount(powerpage, character, "power")
        end
    end
    return ai.dumb(combatstate)
end

--[[
-- An AI for Blood Moon. As a clone of Scarlet Moon, Blood Moon
-- has most of Scarlet Moon's powers. She buffs herself up and then
-- starts kicking ass and taking names.
--]]
function M.bloodmoonai(combatstate)
    local character = combatstateops.activecharacter(combatstate)
    log.info(character.name .. " is using Blood Moon's AI.")
    local powercount = combatstate.choicecounts[character.name].power
    local opponents = combatstateops.activeopponents(combatstate)
    if powercount.current > 0 then
        if not statusops.statusduration(character, statuses.swinging) then
            return ai.decrementcount(
                scarletmoonpowers["s-cords"],
                character,
                "power"
            )
        end
        if not statusops.statusduration(character, statuses.scarletsaber) then
            return ai.decrementcount(
                scarletmoonpowers["s-saber"],
                character,
                "power"
            )
        end
        if not statusops.statusduration(character, statuses.scarletbeam) then
            return ai.decrementcount(
                scarletmoonpowers["s-beam"],
                character,
                "power"
            )
        end
        if not statusops.statusduration(character, statuses["s-armor"]) then
            return ai.decrementcount(
                scarletmoonpowers["s-armor"], 
                character, 
                "power"
            )
        end
    end
    return ai.dumb(combatstate)
end

return M
