--[[
-- Copyright 2016 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]

DEBUG=false
-- When set to true, the game will generate an HTML-only version of the
-- game. Essentially, this generates a transcript for people to peruse
-- without having to play the game itself.
GENERATE_HTML=false
GENERATE_HTML_GENDER="female"

PATH_SEPARATOR = string.sub(package.config, 1, 1)
local status, love = pcall(require, "love")
LOVE=status
if not LOVE then
    love = {}
end

local state = require 'state'
local log = require 'log'
local suit = LOVE and require 'SUIT' or nil

local title = require 'utilities/titlescreen'

-- Make sure the save directory is created before the log tries to write to it.
local success, error = LOVE and love.filesystem.write("touch", "")
assert(success or not LOVE, error)
if LOVE then love.filesystem.remove("touch") end

local pathseparator = string.sub(package.config, 1, 1)
local logfilename = "ScarletMoon.log"
local logfile
if love.filesystem then
    logfile = love.filesystem.getSaveDirectory()
        .. pathseparator
        .. logfilename
else
    logfile = logfilename
end
if LOVE then
    love.filesystem.remove(logfilename)
else
    os.remove(logfilename)
end
log.outfile = logfile
log.level = DEBUG and "debug" or "info"

local error
local errortraceback
local function errorhandler(err)
    error = err
    errortraceback = debug.traceback()
    log.error(err)
    log.error(errortraceback)
end

--[[
-- Returns true if this game is running on Unix, false otherwise.
--]]
local function unix()
    return PATH_SEPARATOR == "/"
end


--[
-- A list of button ids, where the index of the id corresponds to the index
-- of that button's choice in the current page. Populated by `transition`,
-- and used by update to resolve player input.
--]
local choiceIds = {}

--[
-- Transitions to the specified page. Transitioning involves setting the
-- specified state's current page to the specified page, invoking the
-- page's transformations on the state, drawing the GUI, and populating
-- `choiceIds` with the IDs of the player's choices
--
-- uistate -- The current UI state
-- page -- The page to transition to
--]
local function transition(uistate, page)
    local previousPage = uistate.core.page
    uistate.core.page = page
    -- We reference `state.core.page` rather than `page` because a
    -- page's transformer may dynamically change state.core.page (i.e.
    -- the back page does this).
    uistate.core.page.transformer(uistate, previousPage, uistate.core.page)
    local nextPage = uistate.core.page.automatics(uistate)
    if nextPage then
        transition(uistate, nextPage)
    else
        choiceIds = assert(
            uistate.core.page.choices(uistate)._keys,
            "A page's choices table must be ordered (i.e. have a _keys "
            .. "attribute that contains a list of every key in the table.)"
        )
    end
end

local uistate = state.builduistate(
    state.buildgamestate(state.buildconfiguration())
)

--[[
-- Performs the initial setup for the game, including loading game pages,
-- and initializing the uistate.
--]]
function love.load()
    if GENERATE_HTML then
        require 'generate-org-file'.generateOrgMode()
    end
    if LOVE then
        suit.theme.color.normal.fg = {0,0,0}
        suit.theme.color.normal.bg = {.75,.75,.75}
        love.graphics.setBackgroundColor(.9,.9,.9,.9)
    end
    transition(uistate, title.title)
end

--[[
-- Returns true if this game is running on Unix, false otherwise.
--]]
local function unix()
    return PATH_SEPARATOR == "/"
end

local function clear_screen()
    if unix() then
        -- Clear the screen, including the scrollback buffer.
        os.execute("printf \"\\e[3J\\033c\"")
    else
        os.execute("cls")
    end
end

--[[
-- Runs every game tick, refreshes the screen, and responds to any user input.
--]]
function love.update()
    if LOVE then
        uistate.gui.layout:reset(100, 100)
        uistate.gui.layout:padding(10, 10)
    end

    if not LOVE then
        clear_screen()
    end
    local page = uistate.core.page
    local result = xpcall(page.drawgui, errorhandler, uistate)
    assert(
        result,
        "I'm dying. Please, if I ever meant anything to you, "
        .. "give this message:\n\n"
        .. tostring(error) .. "\n"
        .. tostring(errortraceback) .. "\n"
        .. "and this log file "
        .. log.outfile
        .. "\n\nto aka@spankingrpgs.com along with a description of what you were"
        .. " doing when I died. He will know what to do.\n\n"
        .. "Or he won't, and he'll ask you a bunch of inane questions. "
        .. "Guy's not exactly the sharpest tool in the shed, you know what I'm "
        .. "saying?"
    )

    local commands
    if LOVE then
        local command = suit.getPressedKey()
        if not command then
            for _, id in ipairs(choiceIds) do
                if suit.isHit(id) then
                    command = id
                end
            end
        end
        commands = {command}
    else
        local command = string.lower(io.stdin:read())
        if #command == 0 then
            commands = {"return"}
        else
            commands = {}
            for singlecommand in string.gmatch(command, "[^;]+") do
                table.insert(commands, singlecommand)
            end
        end
    end
    for i, command in ipairs(commands) do
        log.info("Command: " .. command)
        uistate.command = command
        if LOVE then
            local ctrlDown = love.keyboard.isDown("rctrl")
                or love.keyboard.isDown("lctrl")
            if command == "return" and ctrlDown then
                love
                    .window
                    .setFullscreen(not love.window.getFullscreen(), "desktop")
            end
        end
        local choices = page.choices(uistate)
        local nextpage = choices[command]
        if nextpage then
            local oldpagename = uistate.core.gamestate.storypagename
            log.info("oldpagename: " .. (oldpagename or "nil"))
            assert(
                xpcall(transition, errorhandler, uistate, nextpage),
                "An error occurred while trying to process your command. Please"
                .. " send the file "
                .. log.outfile
                .. " to sprpgs@gmail.com along with a description of what you were"
                .. " doing when the error occurred. Sorry!"
            )
            local newpagename = uistate.core.gamestate.storypagename
            log.info("newpagename: " .. (newpagename or "nil"))
            if oldpagename ~= newpagename then
                log.info("Story page: " .. (newpagename or "nil"))
            end
        end
        if i < #commands then
            page = uistate.core.page
            result = xpcall(page.drawgui, errorhandler, uistate)
            assert(
                result,
                "I'm dying. Please, if I ever meant anything to you, "
                .. "give this message:\n\n"
                .. tostring(error) .. "\n"
                .. tostring(errortraceback) .. "\n"
                .. "and this log file "
                .. log.outfile
                .. "\n\nto sprpgs@gmail.com along with a description of what you were"
                .. " doing when I died. He will know what to do.\n\n"
                .. "Or he won't, and he'll ask you a bunch of inane questions. "
                .. "Guy's not exactly the sharpest tool in the shed, you know what I'm "
                .. "saying?"
            )
        end
    end
end

function love.draw()
    suit.draw()
end

function love.keypressed(key)
    suit.keypressed(key)
end

function love.textinput(char)
    suit.textinput(char)
end

if not LOVE then
    love.load()
    while true do
        love.update()
    end
end
