--[[
-- Copyright 2017 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]
--
--[[
-- This module contains the atomic predicates that can be used to define predicates
-- in `data/predicates` for use in transitioning automatically to game events
--]]

local predicates = {}

--[[
-- Evaluates to true if the specified character has the specified
-- value for the specified attribute.
--
-- For example, attribute(greg; gender; male) will return true if the character greg is
-- male, and false otherwise.
--
-- character - The character whose attribute needs to be checked
-- attributename - The name of the attribute that should be checked.
--  value - The value that the character's attribute should be compared against
--]]
function predicates.attribute(character, attributename, value)
    return character.attributes[attributename] == value
end

-- [[
-- Evaluates whether or not the player has acquired the specified keyword.
--
-- keywords - All the keywords accumulated by the player
-- keyword - The keyword to check for
-- ]]
function predicates.keyword(keywords, keyword)
    return keywords[keyword]
end

--[[
-- Checks the relative value of the specified statistic for the specified
-- character. For example, statistic;greg;hunger;gt;10 returns true if
-- greg has more than 10 hunger.
--
-- character - The character to be tested
-- statisticname - The name of the statistic to compare
-- comparison - The comparison to perform. This can have one of the following values:
--  gt - Greater than
--  lt - Less than
--  e - Equal
--  gte - Greater than or equal
--  lte - Less than or equal
-- value - The value to compare the character's statistic against
--]]
function predicates.statistic(character, statisticname, comparison, value)
    if comparison == "e" then
        return character.statistics[statisticname] == value
    elseif comparison == "gt" then
        return character.statistics[statisticname] > value
    elseif comparison == "lt" then
        return character.statistics[statisticname] < value
    elseif comparison == "gte" then
        return character.statistics[statisticname] >= value
    elseif comparison == "lte" then
        return character.statistics[statisticname] <= value
    else
        assert(false, "Unknown operation " .. comparison)
    end
end

return predicates
