--[[
-- Copyright 2017 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]

-- [[
-- This module defines the function `load`. `load` takes a gamestate, and
-- loads in all of the M.pages for a game. Story pages are found in
-- the data/story folder.
-- ]]

local M = {}

local storyparser = require 'orgparser'
local log = require 'log'

local pages = require 'pages'
local gamepages = require 'story/storypages'
local scarletfunctions = require 'scarletfunctions'
local scarlettables = require 'scarlettables'

local unpack = scarletfunctions.unpack

local function trim(str)
    return string.gsub(str, "^%s*(.-)%s*$", "%1")
end

M.SKIP_COMBAT_PREFIX = "$$skip$$"

function M.skipcombatname(eventname)
  return M.SKIP_COMBAT_PREFIX .. eventname
end

--[[
-- An operation that populates the configuration
-- with the pages defined in data/story.
--
-- configuration: The table of game configuration to populate
-- filenaem: The name of the file to load, i.e. 'episode1_day2.story'
--
-- returns The table of loaded stories
--]]
function M.load(configuration, filename)
    -- We keep this local to this function to make it easier to test the helper
    -- functions defined in this module.
    -- if filename is nil, then load every storyfile in loveload.
    return LOVE
        and M.loveload(configuration, filename)
        or M.consoleload(configuration, filename)
end

--[[
-- Given a table of events, returns a table of story pages
--
-- events: An ordered table mapping event name to events. See story/stories.md
--  for details on the layout of events
--
-- returns A table mapping event names to story pages
--]]
local function loadevents(events, configuration)
    local stories = {}
    for _, eventname in ipairs(events._keys) do
        local event = events[eventname]
        log.debug("Loading event: " .. eventname)
        if event.allies then -- Only combat events have an allies key.
          local combatpage = M.loadcombatpage(event)
          combatpage.name = eventname
          stories[eventname] = combatpage
          local skipcombat = pages.page(
                function(gamestate)
                    local name = gamestate.characters.pc.combatname
                    return {"pow! smash! kathunk! zap!"}
                end,
                function(gamestate)
                    local state = require 'state'
                    return not combatpage.defeat
                        and state.getstory(configuration, combatpage.victory)
                        or nil
                end,
                function(gamestate)
                    local state = require 'state'
                    local victorystring = "victory! i am invincible!"
                    local defeatstring = "defeat! oh noes!"
                    local victory = state.getstory(
                        configuration, 
                        combatpage.victory
                    )
                    local defeat = state.getstory(
                        configuration,
                        combatpage.defeat
                    )
                    return {
                        [victorystring]=victory,
                        [defeatstring]=defeat,
                        _keys={victorystring, defeatstring},
                    }
                end
          )
          skipcombat.name = M.skipcombatname(eventname)
          stories[skipcombat.name] = skipcombat
        else
            log.debug("Building body for " .. eventname)
            local body = M.buildbody(configuration, event.story)
            local transformations = M.builddo(
                configuration,
                eventname,
                event['do']
            )
            local automatics = M.buildautomatics(
                configuration,
                event.automatics
            )
            local choices = M.buildchoices(configuration, event.choices)
            stories[eventname] = pages.page(
                body,
                automatics,
                choices,
                transformations
            )
            stories[eventname].name = eventname
        end
    end
    return stories
end


--[[
-- Loads events from *.story files into memory.
--
-- storylines: An iterator of lines over a story file
--
-- return An ordered table mapping names to events
--]]
local function parseevents(storylines, filename)
    local events = {_keys={}}
    local errors = {}
    local newevents = storyparser.decode(storylines, filename)
    for _, name in pairs(newevents._keys) do
        local event = newevents[name]
        if events[name] then
            table.insert(errors, "Event " .. name .. " already exists!")
        end
        events[name] = event
        table.insert(events._keys, name)
    end
    assert(#errors == 0, table.concat(errors, "\n"))
    return events
end

--[[
-- An operation that populates the configuration
-- with the pages defined in the specified file using vanilla Lua.
--
-- configuation: The configuration to populate
-- filename: The name of the file to load story events from
--
-- returns A table mapping event names to story pages
--]]
function M.consoleload(configuration, filename)
    assert(filename, "The console doesn't support loading all story files at once.")
    local directory = table.concat({"data", "story"}, PATH_SEPARATOR)
    local storyfilename = table.concat({directory,filename}, PATH_SEPARATOR)
    local events = parseevents(io.lines(storyfilename), storyfilename)
    return loadevents(events, configuration)
end

--[[
-- An operation that populates the configuration
-- with the pages defined in data/story using Love's filesystem utilities.
--
-- configuration: The table of game configuration to populate
-- directory: The directory relative to the game root to search for story files,
-- defaults to "data/story
--
-- returns The table of loaded stories
--]]
function M.loveload(configuration, filename)
    local love = require 'love'
    directory = "data/story"
    local stories = {}
    if filename then 
        directory = directory .. "/" .. filename
        local events = parseevents(love.filesystem.lines(directory), directory)
        return loadevents(events, configuration)
    else
        for _, file in pairs(love.filesystem.getDirectoryItems(directory)) do
            if string.match(file, ".*%.story") then
                local events = parseevents(
                    love.filesystem.lines(directory .. "/" .. file), 
                    directory
                )
                local somestories = loadevents(events, configuration)
                for name, story in pairs(somestories) do
                    stories[name] = story
                end
            end
        end
        return stories
    end
end

--[[
-- Given a table containing the definition of a battle, builds a
-- table with the following keys:
--
-- 1. victory - The name of the M.age to transition to upon victory
-- 2. defeat - The name of the M.age to transition to upon defeat (if any)
-- 3. allies - The list of names of the characters the player is controlling.
-- 4. enemies - The list of names of the characters the enemy is controlling
--
-- Unfortunately, this is one place where the architecture fails a little bit,
-- in that the combat definition are not pages. Instead, they are tables
-- with the following keys:
--  victory - The name of the page to transition to after a victory
--  loss - The name of the page to transition to after losing
--  allies - A list of names of the characters the player will control
--      in battle
--  enemies - A list of the names of the characters the player will battle
--
--  Furthermore, this really should be returning a page. However, the logic
--  to begin combat works on uistate, meaning this function would need to
--  return a uipage. However, M.events are gamepages (i.e. events on
--  gamestate).
--
--  It's ugly, and really should be reworked. One option is to see if we
--  could somehow view combat pages as M.pages with special choices. Then
--  combat pages would all exist in the gamestate.
--]]
function M.loadcombatpage(event)
    return event
end

--[[
-- Given a templated string, returns a
-- function that takes a gamestate, and resolves the templated text into
-- concrete text to display to the user.
--
-- configuration - The configuration state
-- body - The templated string
-- name - The name of the event whose body is being constructed, may be `nil`
--  if building text that isn't the body of an actual event
--
-- returns A function that takes gamestate, and returns a list of strings, where
-- each string represents a paragraph of text
--
--]]
function M.buildbody(configuration, body)
    log.debug("Building body: ", body)
    local templatedtext = M.buildtemplatedbody(configuration, body)
    log.debug("Number of templated text functions: " .. #templatedtext)

    return function(gamestate)
        local statictextlist = {}
        for i,template in ipairs(templatedtext) do
            statictextlist[i] = template(gamestate)
        end
        local statictext = table.concat(statictextlist)

        local paragraphstart = 1
        local paragraphend = string.find(statictext, "\n\n")
        local paragraphs = {}
        local numparagraphs = 0
        while paragraphstart do
            numparagraphs = numparagraphs + 1
            paragraphs[numparagraphs] = string.sub(
                statictext,
                paragraphstart,
                paragraphend and paragraphend - 1
            )
            paragraphstart = paragraphend and paragraphend + 2
            paragraphend = string.find(statictext, "\n\n", paragraphstart)
        end
        return paragraphs
    end
end

-- [[
-- Given templated text, returns an alternating list of static and
-- dynamic text. The static text are strings, the dynamic text are functions
-- that take a gamestate and return a string.
--
-- body: A string encoding the templated M.event
--
-- return: A list functions that take the gamestate
--  and return a string to display
-- ]]
function M.buildtemplatedbody(configuration, body)
    body = body or ""
    local templatedtext = {}
    local numruns = 0
    local textrun = {}
    local runlength = 0
    local numopenbrackets = 0

    assert(
        type(body) == "string",
        "Expected a string, but isn't: " .. scarlettables.string(body)
    )
    for i = 1, #body do
        local character = string.sub(body, i, i)
        if character == "{" then
            numopenbrackets = numopenbrackets + 1
            if numopenbrackets == 1 then
                numruns = numruns + 1
                local text = table.concat(textrun)
                templatedtext[numruns] = function() return text end
                textrun = {}
                runlength = 0
            elseif numopenbrackets > 1 then
                runlength = runlength + 1
                textrun[runlength] = character
            end
        elseif character == "}" and numopenbrackets == 1 then
            numopenbrackets = 0
            numruns = numruns + 1
            templatedtext[numruns] = M.dynamictext(
                configuration,
                table.concat(textrun)
            )
            textrun = {}
            runlength = 0
        elseif character == "}" then
            numopenbrackets = numopenbrackets - 1
            assert(
                numopenbrackets >= 1,
                "Dangling close bracket found at "
                .. string.sub(body, i, i+20)
            )
            runlength = runlength + 1
            textrun[runlength] = character
        else
            runlength = runlength + 1
            textrun[runlength] = character
        end
    end
    if textrun then
        local text = table.concat(textrun)
        templatedtext[numruns+1] = function() return text end
    end
    return templatedtext
end

--[[
-- Given string representing the internals of dynamic text
-- (i.e. the text inside the outermost brackets), returns a function that
-- takes the gamestate and returns a string of static text to display
--
-- configuration: The configuration state
-- textrun: A string representing the function invocation of dynamic text
--
-- return: A function from gamestate to string
--]]
function M.dynamictext(configuration, dynamictext)
    local templatefunction, argumentsParser = M.parsefunction(
        configuration,
        dynamictext,
        configuration.dynamics
    )
    return function(gamestate)
        return templatefunction(argumentsParser(gamestate))(gamestate)
     end
end

--[[
-- Given a string representing a function/operation (i.e. a string of the form
-- "functionname;arg1;arg2;...;argn"), returns the function specified by `functionname`, and
-- a function that takes a gamestate, and returns all the arguments to apply to `functionname`.
--
-- Note: A function's arguments may in fact be a templated string of arbitrary complexity.
-- Therefore, we need the gamestate in order to fully parse a function's arguments.
--
-- This function will generate an assertion error if `functionname` is not defined in the
-- specified function table
--
--
-- configuration: The configuration state
-- functionstring: A string of the form "functionname;arg1;arg2;...;argn"
-- functiontable: The table containing the function defined by `functionname`
--
-- return Two functions. The first is the function specified by `functionname`, the second
-- takes the gamestate and returns all of the arguments to apply to `functionname`
--]]
function M.parsefunction(configuration, functionstring, functiontable)
    local namend = string.find(functionstring, ";")
    local functionname = string.sub(functionstring, 1, namend and namend - 1)
    functionname = string.gsub(functionname, '"', "")

    local templatefunction = functiontable[functionname]
    assert(
        templatefunction,
        "Template function: '" 
            .. functionname 
            .. "' not found when invoking '" 
            .. functionstring .. "'."
    )

    local rawargumentstring = namend
            and string.sub(functionstring, namend and namend+1)
    local templatedarguments = rawargumentstring
        and M.buildtemplatedbody(configuration, rawargumentstring)

    local parseArguments = function(gamestate)
        if not templatedarguments then
            return nil
        end
        local argumentstrings = {}
        for i = 1, #templatedarguments do
            argumentstrings[i] = templatedarguments[i](gamestate)
                or templatedarguments[i]
        end

        local argumentstring = table.concat(argumentstrings)
        local arguments = {}
        local argstart = 1
        local argend = string.find(argumentstring, ";", argstart)
        while argstart do
            local argument = string.sub(
                argumentstring,
                argstart,
                argend and argend - 1
            )
            arguments[#arguments+1] = trim(argument)
            argstart = argend and argend + 1
            argend = string.find(argumentstring, ";", argstart)
        end
        return unpack(arguments)
    end
    return templatefunction, parseArguments
end

--[[
-- Given a list of strings representing operations on gamestate, returns an operation on
-- gamestate that executes each operation in order.
--
-- Note: This transformer is idempotent. If the player has already visited the
-- associated event, then the transformer does nothing.
--
-- configuration: The configuration state
-- name: The name of the event being constructed
-- doblock: A list of strings representing operations to be performed on a gamestate, may
-- be `nil`
--
-- return An operation on gamestate that executes each operation specified in the doblock in
-- order, or the identity operation if the do block is `nil` or empty
--]]
function M.builddo(configuration, name, doblock)
    if not doblock then
        return function() end
    end
    local operations = {}
    for i, operationstring in ipairs(doblock) do
        local operation, argumentParser = M.parsefunction(
            configuration,
            operationstring,
            configuration.operations
        )
        operations[i] = {operation=operation, argumentParser=argumentParser}
    end
    local function dofunction(gamestate)
        if gamestate.visited[name] then
            return
        end
        for _, operation in ipairs(operations) do
            operation.operation(operation.argumentParser(gamestate))(gamestate)
        end
    end
    if GENERATE_HTML then
        return function(gamestate, originalblock)
            if originalblock then
                return doblock
            else
                dofunction(gamestate)
            end
        end
    else
        return dofunction
    end
end

--[[
-- Given an ordered table that maps strings representing predicates on gamestate to event names,
-- and an operation on gamestate, returns a function that takes a gamestate and returns
-- the next event to transition it automatically, or nil if there is no such event.
--
-- configuration: The configuration state
-- automatics: An ordered table of predicate strings to event names, may be nil
--
-- returns A function that takes the gamestate and returns the next page to
-- transition to automatically, if any, if automatics is nil then this returns
-- a noop
--]]
function M.buildautomatics(configuration, automatics)
    if not automatics then
        return function() end
    end
    local predicates = {}
    local pagenames = {}
    for i, predicatestring in ipairs(automatics and automatics._keys or {}) do
        pagenames[i] = automatics[predicatestring]
        local predicate, arguments = M.parsefunction(
            configuration,
            predicatestring,
            configuration.predicates
        )
        predicates[i] = function(gamestate)
            return predicate(arguments(gamestate))(gamestate)
        end
    end
    local function automaticfunction(gamestate)
        for i, predicate in ipairs(predicates) do
            if predicate(gamestate) then
                return gamepages.page(gamestate.configuration, pagenames[i])
            end
        end
        return nil
    end
    if GENERATE_HTML then
        return function(gamestate, originalautomatics)
            if originalautomatics then
              return automatics
            end
            return automaticfunction(gamestate)
        end
    else
        return automaticfunction
    end
end

--[[
-- Given an ordered table that maps dynamic text to event names,
-- and an operation on gamestate, returns a function that takes a gamestate and
-- returns a table that maps Strings (player choices) to pages. If choices is
-- `nil`, then the function returns an empty table.
--
-- configuration: The configuration state
-- choices: An ordered table of dynamic text to event names, may be `nil`
--
-- returns A function that takes a gamestate and returns an ordered table of strings to
-- pages, the strings are choices the player may choose from, and the pages are the story
-- events that each choice transitions to
--]]
function M.buildchoices(configuration, choices)
    if not choices or not next(choices) then
        return function() return {_keys={}} end
    end
    local dynamicchoices = {}
    local pagenames = {}
    for i, choicebody in ipairs(choices._keys) do
        dynamicchoices[i] = M.buildbody(configuration, choicebody)
        pagenames[i] = choices[choicebody]
    end
    return function(gamestate)
        local resolvedchoices = {_keys={}}
        for i, dynamicchoice in ipairs(dynamicchoices) do
            local choicebody = table.concat(dynamicchoice(gamestate), "\n\n")
            choicebody = string.gsub(choicebody, "\\\"", "\"")
            resolvedchoices._keys[i] = choicebody
            if DEBUG then
                assert(
                    pagenames[i],
                    "Missing page name for index "
                        .. i
                        ..
                        " names: "
                        .. scarlettables.string(pagenames)
                )
            end
            resolvedchoices[choicebody] = gamepages.page(
                gamestate.configuration,
                pagenames[i]
            )
        end
        return resolvedchoices
    end
end


return M
