--[[
-- Copyright 2017 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]
--
--[[
-- Defines the page that allows the player to control game options, like
-- difficulty. The options screen has roughly the following structure in
-- the Love GUI:
--
-- Option1: value1  value2  value3
-- Option2: value4  value5  value 6
-- ...
-- (Enter) Confirm (⇐)Back
--
-- Options are navigated using the up and down arrow, while the left and right
-- arrows are used to select the appropriate value.
--
-- The option screen is laid out as follows:
--
-- optionname1: optionvalue1
-- optionname2: optionvalue2
-- ...
-- optionnamen: optionvaluen
--
-- where `optionname` represents the names of the options, while `optionvalue`
-- represents the current values for those M. The currently selected
-- option is highlighted, and the user can select an option with the up and
-- down arrow _keys. When an option is selected, the user can use the right and
-- left arrows to cycle through the possible values for the selected option.
--
--
-- In order to track the player's manipulation of options, the uistate keeps
-- a temporary table called `options`. This table has two keys:
--  selection - The name of the option that is currently selected
--  settings - A table that maps the name of each option to the name of the
--      currently selected option value
--
-- The Console UI looks like the character screen UI:
--
-- (1) optionname1: optionvalue1
-- (2) optionname2: optionname2
-- ...
-- (n) optionnamen: optionvaluen
--
-- Selecting an option (either by number or pattern search) provides a 
-- list of searchable optoins to select from.
--]]

local M = {}

local fonts = LOVE and require 'ui/fonts'
local love = LOVE and require 'love'
local log = require 'log'
local pages = require 'pages'
local scarlettables = require 'scarlettables'
local uiutilities = require 'ui/utilities'
local consoleutils = require 'ui/consoleutils'

local function clearoptions(uistate)
    uistate.options = nil
end

-- [[
-- Builds an options page with the specified transformer
--
-- transformer: An operation on uistate to apply when transitioning to the
-- constructed page
-- nextpage: The page to transition to after the options have been set
-- previouspage: The page to transition to without setting any options,
--  defaults to nextpage if not provided
--
-- return: An options page with the specified transformer
-- ]]
local function buildoptionspage(transformer, nextpage, previouspage)
    local optionspage = pages.page(
        nil, -- No body, the UI is constructed in `drawgui`
        nil, -- No automatics
        function() return M.optionschoicesids(nextpage, previouspage) end,
        transformer
    )
    function optionspage.drawgui(uistate)
        if LOVE then
            M.optionsbody(uistate)
            M.choicesdisplay(uistate)
        else
            M.consoleUI(uistate)
        end
    end
    return optionspage
end


--[[
-- Builds a row for selecting the values of an option. An option row is laid
-- out as follows:
--
-- optionname: optionvalue
--
-- Where `optionnname` is the name of the option, and `optionvalue` is the
-- currently selected value for the option. The row is bold if that option
-- is currently selected.
--
-- params:
--  optionname - The name of the option
--  optionvalue - The name of the currently selected value
--  font - The font to use to render the row
--  suit - The UI toolkit to use, following the interface of SUIT
--  width - The width of the game screen
--  index - The number of the option we're looking at (starting at 1 at the
--      top of the screen and counting up as we go down the screen)
--]]
local function buildoptionrow(optionname, valuename, font, suit)
    local optiontext = optionname .. ": " .. "<- " .. valuename .. " ->"
    local textlength = fonts.width(font, optiontext)
    local textheight = font:getHeight()
    suit.Label(
        optiontext,
        {
            id=optionname,
            font=font
        },
        suit.layout:row(textlength, textheight)
    )
end

--[[
-- Displays the table of currently selected options to the user. The screen is
-- laid out as follows:
--
-- optionname1: optionvalue1
-- optionname2: optionvalue2
-- ...
-- optionnamen: optionvaluen
--
-- where `optionname` represents the names of the options, while `optionvalue`
-- represents the current values for those M. The currently selected
-- option is highlighted.
--
-- uistate: The current state of the UI
--  In addition to the standard mappings, this function assumes the following
--  additional mappings:
--      options: A table containing the current state of the options
--          screen with the following mappings:
--              selected: The name of the currently selected option. Defaults
--                  to the first value in
--                  uistate.core.gamestate.configuration.options._keys
--              settings: A table of the currently selected values for each
--                  option, keyed on option name. Defaults to the current
--                  setting for each option in `gamestate.options`.
--]]
function M.optionsbody(uistate)
    local suit = uistate.gui
    local options = uistate.core.gamestate.options
    local optionstable = uistate.core.gamestate.configuration.options
    local width, height = love.graphics.getDimensions()
    local numoptions = #optionstable._keys
    local optionfont = fonts.button(options.buttonfont)
    local fontsize = tonumber(options.buttonfont)
    local startingheight = 64 * fonts.title(options.titlefont):getLineHeight()
    local maxlineheight = math.floor((height - startingheight) / numoptions)

    -- May need to tweak the starting position.
    suit.layout:push(width/2, startingheight)

    while optionfont:getLineHeight() > maxlineheight do
        fontsize = fontsize - 1
        optionfont = fonts.buildbutton(fontsize)
    end
    local boldoptionfont = fonts.buildbutton(fontsize, fonts.BOLD)

    local selectedoption = uistate.options.selected

    local currentsettings = uistate.options.settings

    for _, optionname in ipairs(optionstable._keys) do
        local font = selectedoption == optionname
            and boldoptionfont
            or optionfont
        local valuename = currentsettings[optionname]
        buildoptionrow(optionname, valuename, font, suit)
    end
    suit.layout:pop()
end

--[[
-- Constructs the labels that display the user's choices. The
-- user has the following choices:
--  (<==)Back: Go back to the previous screen without applying any changes
--  (Enter) Apply: Go back to the previous screen, and apply the newly selected
--      option values.
--  (Up/Down Arrows): Select the option either above or below the currently
--      selected option.
--  (Right/Left Arrows): Select the option value listed before or after the
--      currently selected option value.
--
-- uistate: The current state of the UI
--]]
function M.choicesdisplay(uistate)
    local suit = uistate.gui
    local font = fonts.button(uistate.core.gamestate.options.buttonfont)
    suit.Button(
        uiutilities.up,
        {
            id="up",
            font=font
        },
        suit.layout:row(fonts.width(font, uiutilities.up), font:getHeight())
    )
    suit.Button(
        uiutilities.down,
        {
            id="down",
            font=font
        },
        suit.layout:col(
            fonts.width(font, uiutilities.down),
            font:getHeight()
        )
    )
    local applytext = "(Enter) Apply"
    suit.Button(
        applytext,
        {
            id="return",
            font=font
        },
        suit.layout:col(fonts.width(font, applytext), font:getHeight())
    )
    local backtext = "(<==)Cancel"
    suit.Button(
        backtext,
        {
            id="backspace",
            font=font
        },
        suit.layout:col(fonts.width(font, backtext), font:getHeight())
    )
end


-- [[
-- A helper function that updates the currently selected option based on the
-- specified delta.
--
-- uistate: The current state of the UI
--  In addition to the standard mappings, this function assumes the following
--  additional mappings:
--      options: A table containing the current state of the options
--          screen with the following mappings:
--              selected: The name of the currently selected option.
--
-- delta: Either 1 if we want to go forward or -1 if we want
--  to go backwards through the list of options
-- ]]
local function updowntransformer(uistate, delta)
    local optionstable = uistate.core.gamestate.configuration.options
    local numoptions = #optionstable._keys
    local currentindex = scarlettables.reverseLookup(
        uistate.options.selected,
        optionstable._keys
    )
    local newindex
    if delta > 0 then
        newindex = currentindex == numoptions and 1 or (currentindex + 1)
    elseif delta < 0 then
        newindex = currentindex == 1 and numoptions or (currentindex - 1)
    end
    uistate.options.selected = optionstable._keys[newindex]
end

-- [[
-- A helper function that updates the value of the currently selected option
-- based on the specified delta.
--
-- uistate: The current state of the UI
--  In addition to the standard mappings, this function assumes the following
--  additional mappings:
--      options: A table containing the current state of the options
--          screen with the following mappings:
--              selected: The name of the currently selected option.
--              settings: A table of the currently selected value names for
--                  each option, keyed on option name.
--
-- delta: Either 1 if we want to go forward or -1 if we want
--  to go backwards through the list of option values
-- ]]
local function leftrightransformer(uistate, delta)
    local optionstable = uistate.core.gamestate.configuration.options
    local selectedoptionname = uistate.options.selected
    local optionvalues = optionstable[selectedoptionname]
    local numvalues = #optionvalues
    local selectedvaluename = tostring(uistate.options.settings[selectedoptionname])
    local currentindex
    for i, value in ipairs(optionvalues) do
        if value.name == selectedvaluename then
            currentindex = i
            break
        end
    end
    local newindex
    if delta < 0 then
        newindex = currentindex == 1 and numvalues or currentindex - 1
    elseif delta > 0 then
        newindex = currentindex == numvalues and 1 or currentindex + 1
    end
    uistate.options.settings[selectedoptionname] = optionvalues[newindex].name
end

-- [[
-- Applies the chosen options to the current gamestate.
--
-- uistate: The current state of the UI
--  In addition to the standard mappings, this function assumes the following
--  additional mappings:
--      options: A table containing the current state of the options
--          screen with the following mappings:
--              selected: The name of the currently selected option. Defaults
--                  to the first value in
--                  uistate.core.gamestate.configuration.options._keys
--              settings: A table of the currently selected values for each
--                  option, keyed on option name. Defaults to the current
--                  setting for each option in `gamestate.options`.
--
-- ]]
local function applyTransformer(uistate)
    local gamestate = uistate.core.gamestate
    local optionsconfiguration = gamestate.configuration.options
    for optionname, newvaluename in pairs(uistate.options.settings) do
        local oldvaluename = gamestate.options[optionname]
        local optionvalues = optionsconfiguration[optionname]
        if newvaluename ~= oldvaluename then
            local oldoption
            local newoption
            for _, value in ipairs(optionvalues) do
                if value.name == oldvaluename then
                    oldoption = value
                elseif value.name == newvaluename then
                    newoption = value
                end
            end
            if oldoption then
                oldoption.remove(gamestate)
            end
            newoption.apply(gamestate)
            gamestate.options[optionname] = newvaluename
            log.info(
                "Set option: "
                .. optionname
                .. " to "
                .. gamestate.options[optionname]
            )
        end
    end
    clearoptions(uistate)
end


-- [[
-- Constructs a function that returns the table of choices ids to pages that
-- represent the result of
-- making each choice. The user has the following choices:
--
--  (<==)Back: Go back to the previous screen without applying any changes
--  (Enter) Apply: Go back to the previous screen, and apply the newly selected
--      option values.
--  (Up/Down Arrows): Select the option either above or below the currently
--      selected option.
--  (Right/Left Arrows): Select the option value listed before or after the
--      currently selected option value.
--
-- nextpage: The page to transition after the options have been set
-- previouspage: The page to transition to when backing out of the options
-- screen, defaults to `nextpage` if not provided
--
-- return: A table of UI pages keyed on button ids
-- ]]
function M.optionschoicesids(nextpage, previouspage)
    if not previouspage then
        previouspage = nextpage
    end
    local upTransformer = function(uistate)
        updowntransformer(uistate, -1)
    end
    local upPage = buildoptionspage(upTransformer, nextpage, previouspage)

    local downTransformer = function(uistate)
        updowntransformer(uistate, 1)
    end
    local downPage = buildoptionspage(downTransformer, nextpage, previouspage)

    local leftTransformer = function(uistate)
        leftrightransformer(uistate, -1)
    end
    local leftPage = buildoptionspage(leftTransformer, nextpage, previouspage)

    local rightTransformer = function(uistate)
        leftrightransformer(uistate, 1)
    end
    local rightPage = buildoptionspage(rightTransformer, nextpage, previouspage)

    local backPage = pages.page(
        nil,
        function() return previouspage end,
        nil,
        clearoptions
    )

    local applyPage = pages.page(
        nil,
        function() return nextpage end,
        nil,
        applyTransformer
    )

    local backkey = LOVE and "backspace" or "c"
    return {
        up=upPage,
        down=downPage,
        left=leftPage,
        right=rightPage,
        [backkey]=backPage,
        ["return"]=applyPage,
        _keys={
            "up",
            "down",
            "left",
            "right",
            backkey,
            "return"
        }
    }
end

--[[
-- Given an operation on uistate, returns a transformer that applies that
-- operation, and sets up the uipage page for handling options:
--  1. It makes the options backable
--  2. It initializes the uistate.settings table as follows:
--      1. It uses `configuration.options` to determine which option should
--          be selecte first (it defaults to the first option in
--          `configuration.options`)
--      2. For each `optionname` it initializes
--          `configuration.settings[optionname]` to the value currently
--          selected in `gamestate.options` (i.e. the currently selected value)
--
-- transformer: An optional operation on uistate to apply before making the
-- options page backable
--
-- return: A transformer that applies the passed in transformer and makes the
-- options page backable
--]]
function M.optionstransformer(transformer)
    return function(uistate)
        if transformer then
            transformer(uistate)
        end
        local gamestate = uistate.core.gamestate
        local configuration = gamestate.configuration
        uistate.options = {
            selected=configuration.options._keys[1],
            settings={}
        }
        for _, optionname in ipairs(configuration.options._keys) do
            uistate.options.settings[optionname] = gamestate.options[optionname]
        end
        uistate.options.selected = configuration.options._keys[1]
    end
end

--[[
-- Displays the options screen on the console
--
-- options: The configured table of options. See utilities/options.md for
--  the format of this table
-- chosenoptions: A table mapping option names to the option values (strings)
--  that the player has chosen at this point
--]]
local function displayoptionsconsole(options, chosenoptions, issearching)
    local lines = {}
    for i, optionname in ipairs(options._keys) do
        local option = options[optionname]
        local chosen = chosenoptions[optionname]
        local line = string.format("(%d) %s: %s", i, optionname, chosen)
        table.insert(lines, line)
    end

    local title = "------------ Options --------------"
    local commands = table.concat(
        {
            "(Enter) Done",
            issearching and "" or "(C)ancel"
        },
        "\t"
    )
    print(table.concat(
        {
            title,
            "",
            table.concat(lines, "\n"), 
            "",
            commands
        }, 
        "\n"
    ))
end

--[[
-- Provides a list of possible values for the chosen option, asking the
-- player to select the one they want.
--
-- choice: Table with two keys: 
--  id - containing the name of the option, and
--  values - A list of all possible values for the option
-- chosenoptions: A table mapping option names to the value currently chosen
--]]
local function changeoptionconsole(choice, chosenoptions)
    if #choice.values == 1 then
        chosenoptions[choice.id] = choice.values[1].name
        return true
    end
    local optionnames = {}
    for i, option in ipairs(choice.values) do
        optionnames[option.name] = option
    end
    local lines = {}
    for i, value in ipairs(choice.values) do
        table.insert(lines, string.format("(%d) %s", i, value.name))
    end
    local heading = string.format(
        "------------ %s --------------",
        choice.search or choice.id
    )
    print(table.concat({heading, table.concat(lines, "\n")}, "\n"))
    local command = io.stdin:read()
    if #command == 0 then
        return false
    else
        local commandnumber = tonumber(command)
        if commandnumber then
            local value = choice.values[commandnumber]
            if value then
                chosenoptions[choice.id] = value.name
                return true
            end
        elseif optionnames[command] then
            chosenoptions[choice.id] = optionnames[command].name
            return true
        else
            return consoleutils.patternmatch(
                choice,
                command, 
                function(value) return value.name end,
                function(achoice)
                     return changeoptionconsole(achoice, chosenoptions)
                end
            )
        end
    end
    return false
end

--[[
-- Processes user input for selecting options
--
-- command: The user string provided by the user
-- options: The ordered table of all the configured options
-- chosenoptions: The table of option values currently selected by the user
--]]
local function processconsoleinput(command, options, chosenoptions)
    if input then
        command = io.stdin:read()
    end
    if command and #command > 0 then
        local commandnumber = tonumber(command)
        if commandnumber then
            local optionname = options._keys[commandnumber]
            if optionname then
                local choice = {id=optionname, values=options[optionname]}
                changeoptionconsole(choice, chosenoptions)
            end
        elseif options[command] then
            local choice = {id=command, values=options[command]}
            changeoptionconsole(choice, chosenoptions)
        end
    end
    displayoptionsconsole(options, chosenoptions, command and command.search)
end

--[[
-- Builds a console version of the UI.
--]]
function M.consoleUI(uistate)
    local gamestate = uistate.core.gamestate
    local options = gamestate.configuration.options
    local chosenoptions = uistate.options.settings
    processconsoleinput(uistate.command, options, chosenoptions)
end

--[[
-- Builds a page for displaying game options, and processing the user's
-- changes.
--
-- nextpage: The page to transition to after the options have been set
-- transformer: The transformer to apply when transitioning to this page,
--  optional
-- previouspage: The page to transition to after the options have been set,
--  defaults to nextpage if not provided
--]]
function M.options(nextpage, transformer, previouspage)
    return buildoptionspage(
        M.optionstransformer(transformer),
        nextpage,
        previouspage
    )
end

return M

