--[[
-- Copyright 2017 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- uiutilitiesERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]
--
--[[
-- A module that contains functions for saving and loading a game.
--
-- We want to save the bare minimum necessary to derive the gamestate, so that
-- in-progress games can pick up changes to the configuration:
--
-- 1. For characters, we only save player characters, and even then only the
--  state the player can affect.
-- 2. We don't save the configuration.
-- 3. We save the keywords
-- 4. We only save the names of the value for each option.
--
-- Savegames are tables serialized to JSON, defined as follows:
-- {
--  characters={
--      charname={
--          powers={"power1name", "power2name"},
--          clothing={slot1="clothing1name", slot2="clothing2name"},
--          statuses={{statusname=duration}, {statusname2=duration2}},
--          statistics={statisticname=number, statistic2name=number}
--      }
--  },
--  keywords={keyword=true, keyword2=true},
--  options={option1="optionvaluename", option2="optionvaluename"},
--  lossNumber=number,
--  storypagename=pagename
-- }
--
--]]

local M = {}

local json = require 'json-parser'
local log = require 'log'
local state = require 'state'

--[[
-- Saves the specified gamestate to the specified file in JSON format.
--
--  gamestate - The current state of the game
--  savename - The name of the file to save the game to
--
--  return success if the operation was successful
--  return message Error message if the operation was not successful
--]]
function M.save(gamestate, savename)
    local serializedgamestate = json.encode(M.serializablegamestate(gamestate))
    if LOVE then
        local love = require 'love'
        assert(savename ~= ".sav")
        return love.filesystem.write(savename, serializedgamestate)
    else
        local savepath = table.concat(
            {"saves", savename},
            PATH_SEPARATOR
        )
        local savefile = assert(io.open(savepath, "w"))
        local success, error = savefile:write(serializedgamestate)
        savefile:close()
        return success, error
    end
end

--[[
-- Given the gamestate, returns a new table that contains only the information
-- we want to save.
--
-- We don't want to blindly save the entire gamestate because otherwise people
-- won't be able to get any tweaks to the configuration, effectively making
-- savegames invalid across versions. Instead, we only want to save the bare
-- minimum:
--
-- 1. For characters, we only save player characters, and even then only the
--  state the player can affect.
-- 2. We save the keywords
-- 3. We only save the names of the currently selected value for each option.
--
-- gamestate - The state to simplify for saving purposes
--
-- returns A table that contains all and only the information we need to save
--]]
function M.serializablegamestate(gamestate)
    local serializable = {}
    serializable.keywords = gamestate.keywords
    serializable.options = {}
    serializable.options = gamestate.options
    serializable.characters = M.serializecharacters(gamestate.characters)
    serializable.lossNumber = gamestate.lossNumber
    serializable.storypagename = gamestate.storypagename
    serializable.music = gamestate.music
    serializable.gametext = gamestate.gametext
    serializable.visited = gamestate.visited
    if gamestate.originaloutfit then
        serializable.originaloutfit = {}
        for slot, outfit in pairs(gamestate.originaloutfit) do
            serializable.originaloutfit[slot] = outfit.name
        end
    end
    return serializable
end

--[[
-- Given a table of characters, returns a table containing the bare minimum
-- of information necessary to save characters controlled by the player.
--
-- characters - The characters to get ready for serialization
--
-- return A table keyed on character name containing the information that
--  needs to be saved for each player character
-- ]]
function M.serializecharacters(characters)
    local serializable = {}
    for charname, character in pairs(characters) do
        if character.pc then
            local simplechar = {}
            simplechar.powers = {}
            for _, powername in pairs(character.powers._keys) do
                table.insert(simplechar.powers, powername)
            end
            simplechar.clothing = {}
            log.info("Saving character " .. character.name)
            for _, slot in pairs(character.clothing._keys) do
                simplechar.clothing[slot] = character.clothing[slot].name
            end
            simplechar.statuses = {}
            for i, status in pairs(character.statuses) do
                simplechar.statuses[i]={[status.status.name]=status.duration}
            end
            simplechar.statistics = {}
            for _, statistic in pairs(character.statistics._keys) do
                simplechar.statistics[statistic] =
                    character.statistics[statistic]
            end
            simplechar.printedname = character.printedname
            simplechar.combatname = character.combatname
            simplechar.concisename = character.concisename
            simplechar.attributes = {}
            for componentname, component in pairs(character.attributes) do
                simplechar.attributes[componentname] = component
            end
            simplechar.bummarks = character.bummarks
            serializable[charname] = simplechar
        end
    end
    return serializable
end

local function loveload(savename, configuration)
    local love = require 'love'
    local savegamestring, errormessage = love.filesystem.read(savename)
    if savegamestring then
        return M.buildgamestate(json.decode(savegamestring), configuration)
    else
        log.error("Failed to load " .. savename .. " because: " .. errormessage)
        return nil, errormessage
    end
end

local function consoleload(savename, configuration)
    local savepath = "saves" .. PATH_SEPARATOR .. savename
    local savefile, error = io.open(savepath, "r")
    if not savefile then
        return savefile, error
    end
    local savegamestring, errormessage = savefile:read()
    savefile:close()
    if savegamestring then
        return M.buildgamestate(json.decode(savegamestring), configuration)
    else
        log.error("Failed to load " .. savename .. " because: " .. errormessage)
        return nil, errormessage
    end
end

--[[
-- Loads the data stored in the specified save file into the specified state.
--
-- savename - The name of the savefile to load
-- configuration - The game's configuration
--
-- returns the loaded gamestate, or nil and an error message if we failed to
--  read the file
--]]
function M.load(savename, configuration)
    assert(configuration, "Forgot to pass in the game's  configuration.")
    if not string.match(savename, ".*%.sav$") then
        savename = savename .. ".sav"
    end
    if LOVE then
        return loveload(savename, configuration)
    else
        return consoleload(savename, configuration)
    end
end

local function lookupoption(options, valuename)
    for _, option in ipairs(options) do
        if option.name == valuename then
            return option
        end
    end
    return nil
end

--[[
-- Given a table containing the state saved to the savefile, and the current
-- game's configuration, returns a gamestate to resume play from.
--
-- loadedstate - A table of the minimal information written to the save files
-- configuration - The game's current configuration
--
-- return The state of the game to resume
--]]
function M.buildgamestate(loadedstate, configuration)
    local gamestate = state.buildgamestate(configuration)
    gamestate.keywords = loadedstate.keywords
    gamestate.lossNumber = loadedstate.lossNumber
    gamestate.storypagename = loadedstate.storypagename
    if (gamestate.storypagename:find("blood_moon", 1, true) == 1) then
        gamestate.storypagename = "ep6d2_" .. gamestate.storypagename
    end
    gamestate.gametext = assert(
        loadedstate.gametext,
        "This save file is not compatible!"
    )
    gamestate.visited = loadedstate.visited
    local gameoptions = gamestate.options
    for name, valuename in pairs(loadedstate.options) do
        if not string.match(name, "font") or LOVE then
            local options = configuration.options[name]
            local option = lookupoption(options, string.lower(valuename))
            gamestate.options[name] = string.lower(valuename)
            if option then
                option.apply(gamestate)
            else
                log.warn(
                    "Unknown option value: '" .. valuename .. "' for option: " .. name
                )
            end
        end
    end
    if loadedstate.originaloutfit then
        local originaloutfit = {}
        local scarlettables = require 'scarlettables'
        for slot, outfitname in pairs(loadedstate.originaloutfit) do
            local outfit = configuration.items[outfitname]
            assert(outfit, "Unknown outfit: '" .. outfitname .. "'")
            originaloutfit[slot] = outfit
        end
        gamestate.originaloutfit = originaloutfit
    end
    M.loadcharacters(gamestate, loadedstate, configuration)
    return gamestate
end

--[[
-- Sometimes, I change a power's name. However, this means that
-- people may not be able to load old save files. Since having
-- backwards compatible saves is important to me, this function
-- handles transforming old power names into new power names.
--]]
local function backwardscompatibility(powername)
    if powername == "noodle whack" then
        return "noodle_whack"
    end
    if powername == "noodle spank" then
        return "noodle_spank"
    end
    return powername
end

--[[
-- An operation on `gamestate` that loads the characters defined in
-- `loadedstate` into the specified `gamestate`.
--
-- gamestate - The gamestate to load the characters into
-- loadedstate - The saved state to load the characters from
-- configuration - The current game configuration
-- ]]
function M.loadcharacters(gamestate, loadedstate, configuration)
    for name, characterdata in pairs(loadedstate.characters) do
        local character = assert(
            gamestate.characters[name],
            "Unknown character: " .. "'" .. name .. "'"
        )

        -- Recall that statistics are ordered tables, but we don't
        -- bother saving the order in the savestate.
        for statname, value in pairs(characterdata.statistics) do
            character.statistics[statname] = value
        end

        character.powers = {_keys={}}
        local hashealbear = false
        if name == "princess" then
            for _, powername in pairs(characterdata.powers) do
                hashealbear = hashealbear or powername == "healbear"
            end
            if not hashealbear then
                table.insert(characterdata.powers, "healbear")
            end
        end
        for _, powername in pairs(characterdata.powers) do
            -- We've renamed scarlet focus to s-focus to line
            -- up with the other powers in Scarlet Moon's arsenal.
            if powername == "scarlet focus" then
                powername = "s-focus"
            end
            -- We've consolidated s-fists and s-feet into one power, so 
            -- we don't want to load s-feet anymore.
            if powername ~= "s-feet" then
                powername = backwardscompatibility(powername)
                character.powers[powername] = assert(
                    configuration.powers[powername],
                    "Unknown power: '" .. powername .. "'"
                )
                table.insert(character.powers._keys, powername)
            end
        end

        for slot, clothingname in pairs(characterdata.clothing) do
            assert(
                configuration.items[clothingname],
                "Unknown item: " .. clothingname
            )
            character.clothing[slot] = configuration.items[clothingname]
        end

        for _, status in pairs(characterdata.statuses) do
            local statusname, duration = next(status)
            table.insert(
                character.statuses,
                {status=configuration.statuses[statusname], duration=duration}
            )
        end
        character.combatname = characterdata.combatname
        if character.name == "pc" then
            local episode2 =
                string.sub(loadedstate.storypagename, 1, string.len("ep2"))
                == "ep2"
            if episode2 and character.combatname ~= "Scarlet Moon" then
                character.combatname = "Scarlet Moon"
            end
        end
        character.concisename = characterdata.concisename
        if not character.concisename then
            character.concisename = string.sub(character.combatname, 1, 2)
        end
        character.printedname = characterdata.printedname
        local bummarks = characterdata.bummarks
        -- Added for backwards compatibility. Originally, character.bummarks
        -- was a semi-colon separated string, rather than a list of strings.
        if type(bummarks) == "string" then
          character.bummarks = {}
          for mark in string.gmatch(bummarks, "[^;]+") do
            table.insert(character.bummarks, mark)
          end
        else
          character.bummarks = bummarks
        end
        for componentname, component in pairs(characterdata.attributes) do
            if component == "buns" then
                component = "bun"
            end
            character.attributes[componentname] = component
        end
        -- For backwards compatibility with savefiles, hair length was added later.
        if not character.hairlength then
            character.hairlength = "short"
        end
    end
end

return M
