--[[
-- Copyright 2016 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
-- ]]
--
-- [[
-- Defines the events for saving and loading the game. The only publicly
-- available function is `M.saveload` down at the bottom of the file.
-- Everything else is a helper function to help build the saveload page.
-- ]]
local M = {}

local buttons = LOVE and require 'ui/buttons'
local fonts = LOVE and require 'ui/fonts'
local gamesaveload = require 'utilities/saveload'
local love = LOVE and require 'love'
local pages = require 'pages'
local uiutilities = require 'ui/utilities'
local log = require 'log'


--[[ 
-- Builds the UI for games using the Love engine.
--
-- uistate - The current state of the game
-- options - The table of game options
-- visiblesaves - A list of savegames currently displayed on the screen
-- textinput - The table that the `Input` widgit should populate. It should
--  include the key:
--      text: The text currently in the input box
-- gui - The SUIT-like toolkit used to build the GUI
--]]
local function loveui(uistate, options, visiblesaves, textinput, gui)
    local savename = textinput.text
    -- Build the input field for providing a game to save/load
    local titlefont = fonts.title(options.titlefont)
    local savenamewidth = math.max(
        titlefont:getWidth(savename) + 5,
        fonts.singleCharacterWidth(titlefont) * 16
    )
    local screenwidth = love.graphics.getDimensions()
    gui.layout:push(screenwidth / 4, 5)
    local locationtext = "Save location:"
        .. love.filesystem.getSaveDirectory()
    local buttonfont = fonts.button(options.buttonfont)
    gui.Label(
        locationtext,
        {font = fonts.body(options.bodyfont), id = "location"},
        gui.layout:row(
            buttonfont:getWidth(locationtext) + 5,
            buttonfont:getHeight()
        )
    )
    local labeltext = "Save File:"
    gui.Label(
        labeltext,
        {font=titlefont, align="left"},
        gui.layout:row(
            titlefont:getWidth(labeltext) + 5,
            titlefont:getHeight()
        )
    )
    local x, y, w, h = gui.layout:col(savenamewidth, titlefont:getHeight())
    textinput.cursor = string.len(textinput.text) + 1
    gui.Input(textinput, {id = "saveinput", font = titlefont}, x, y, w, h)
    gui.layout:pop()
    -- Give the save label keyboard focus
    if not uistate.redrawsave then
        gui.setHit("saveinput")
        gui.grabKeyboardFocus("saveinput")
        uistate.redrawsave = true
    end

    -- Build list of save games.
    local scarlettables = require 'scarlettables'
    if visiblesaves[1] then
        gui.layout:push(x + w + fonts.singleCharacterWidth(titlefont), y)
        local maxwidth = savenamewidth
        for _, name in pairs(visiblesaves) do
            local width = titlefont:getWidth(name) + 5
            if maxwidth < width then
                maxwidth = width
            end
        end
        local boldbuttonfont = fonts.buttonBold(options.buttonfont)
        gui.Label(
            visiblesaves[1],
            {font=boldbuttonfont},
            gui.layout:col(maxwidth, boldbuttonfont:getHeight())
        )
        for i=2, #visiblesaves do
            gui.Label(
                visiblesaves[i],
                {font=buttonfont},
                gui.layout:row(maxwidth, buttonfont:getHeight())
            )
        end
        gui.layout:pop()
    end
end

--[[
-- Builds a console UI version of the save screen. Like the 
-- other utility UI's, the player selects from a list of saves,
-- they can also search by typing part of a save screen. Once
-- they select a save, they can save the game by pressing (s)
-- and load it by pressing (l).
--
-- uistate: The current state of the game
--  This function creats an additional mapping on uistate: saves,
--  which contains a list of all the save files the player currently
--  has.
-- allsaves: The list of all saves the plaeyr currently has.
-- command: The command provided by the player, or nil if none
--  has been provided yet
--]] 
local function consoleui(uistate, allsaves, command)
    local commandnumber = tonumber(command)
    local visiblesaves = {}
    if not command or command == "return" then 
        visiblesaves = allsaves
    else
        for i, save in ipairs(allsaves) do
            if i == commandnumber or save == command then
                table.insert(visiblesaves, save)
                break
            end
        end
        if #visiblesaves == 0 then
            local escaped = string.gsub(command, "%-", "%%-")
            local escaped = string.gsub(escaped, "%.", "%%.")
            local pattern = ".*" .. escaped .. ".*"
            for _, save in ipairs(allsaves) do
                if string.match(save, pattern) then
                    table.insert(visiblesaves, save)
                end
            end
        end
        if #visiblesaves == 0 and #command > 0 then
            table.insert(visiblesaves, command)
        end
    end
    local currentsave = visiblesaves[1] or ""
    local lines = {}
    for i, save in ipairs(visiblesaves) do
        table.insert(lines, string.format("(%d) %s", i, save))
    end
    uistate.save.textinput.text = currentsave
    uistate.visiblesaves = visiblesaves
    print("------------ Save/Load --------------")
    print(string.format(
        "------------ Selected: %s --------------",
        currentsave
    ))
    print(table.concat(lines, "\n"))
    print(string.format(
        "(S)ave %s\t(L)oad %s\t",
         currentsave,
         currentsave
    )) 
    print("--------------------------------------------------------------------------------")
    local directions = "Enter number, name or search string to select a different \n"
        .. "save, press Enter to cancel:"
    print(directions)
end

--[[
-- Builds the widgets that display the save games.
--
-- There are two widgets next to each other: A label that displays the
-- currently selected savename, and a list of available saves next to it.
--
-- uistate - The current state of the game
-- currentsave - The index of the currently selected savegame
-- textinput - The table that the `Input` widgit should populate. It should
--  include the key:
--      text: The text currently in the input box
-- gui - The SUIT-like toolkit used to build the GUI
-- allsaves - The list of all save games the player currently has
--]]
local function savedisplay(uistate, currentsave, textinput, gui, allsaves)
    if not textinput.text then
        textinput.text = "save"
    end
    local options = uistate.core.gamestate.options
    local savename = textinput.text
    local visiblesaves = {allsaves[currentsave]}
    for i=1,7 do
        if allsaves[currentsave+i] then
            table.insert(visiblesaves, allsaves[currentsave+i])
        end
    end
    if LOVE then
        loveui(uistate, options, visiblesaves, textinput, gui)
    else
        consoleui(uistate, allsaves, uistate.command)
    end
end

local function buildsavechoices(uistate, cansave)
    local save = "(F5)Save"
    local load = "(F9)Load"
    local back = "(Esc)Cancel"
    local screenwidth, screenheight = love.graphics.getDimensions()
    uistate.gui.layout:push(screenwidth / 3, 2 * screenheight / 3)
    local buttonfont = fonts.button(uistate.core.gamestate.options.buttonfont)
    if cansave then
        buttons.buildButton(
            uistate,
            save,
            "f5",
            function(suit)
                return suit.layout:row(
                    fonts.width(buttonfont, save),
                    buttonfont:getHeight()
                )
            end
        )
    end
    buttons.buildButton(
        uistate,
        load,
        "f9",
        function(suit)
            if cansave then
                return suit.layout:col(
                    fonts.width(buttonfont, save),
                    buttonfont:getHeight()
                )
            else
                return suit.layout:row(
                    fonts.width(buttonfont, save),
                    buttonfont:getHeight()
                )
            end
        end
    )
    buttons.buildButton(
        uistate,
        back,
        "escape",
        function(suit)
            return suit.layout:col(
                fonts.width(buttonfont, back),
                buttonfont:getHeight()
            )
        end
    )
    uistate.gui.layout:pop()
end

--[[
-- Returns a list of all the currently known save files.
--]] 
local function find_saves()
    log.info("Looking for saves.")
    local love = require 'love'
    local files = love.filesystem.getDirectoryItems("")
    local savefiles = {}
    for _, file in ipairs(files) do
        if string.match(file, ".*%.sav") then
            table.insert(savefiles, file)
        end
    end
    return savefiles
end

-- [[
-- Builds the main UI for saving and loading the game. The body
-- consists of an input for providing a savename, as well as a
-- list of savegames that can be scrolled through and selected. Note that
-- scrolling through savegames overwrites any edits made to the input field
-- with the name of the selected save.
--
-- uistate: The current state of the UI. In addition to the standard mappings,
--  this function expects the following save-specific attributes:
--      1. allsaves - A list of all save files that currently exist
--      2. save - A table with the following keys:
--          1. newindex - The save file index that was just selected, defaults
--              to 1
--          2. textinput - A table containing a single key `text` that contains
--              the text currently in the input box
--          This operation adds the following mappings:
--              1. oldsave - The index of the previously selected save file
--              2. newsave - The index of the newly selected save file
-- cansave: Whether or not saving is enabled
-- ]]
local function savegui(uistate, cansave)
    local allsaves = LOVE and find_saves() or uistate.allsaves
    local gui = uistate.gui
    local savestate = uistate.save

    if LOVE then 
        gui.layout:push(love.graphics.getDimensions()/2, 0)
    end
    local visiblesaves = uistate.visiblesaves or allsaves
    savedisplay(uistate, savestate.newsave or 1, savestate.textinput, gui, visiblesaves)
    if LOVE then
        local saveup = "saveup"
        local savedown = "savedown"
        local buttonfont = fonts.button(uistate.core.gamestate.options.buttonfont)
        gui.Button(
            uiutilities.up,
            {font = buttonfont, id = saveup },
            gui.layout:col(
                buttonfont:getWidth(uiutilities.up) + 5,
                buttonfont:getHeight()
            )
        )
        gui.Button(
            uiutilities.down,
            {font = buttonfont, id = savedown },
            gui.layout:col(
                buttonfont:getWidth(uiutilities.down) + 5,
                buttonfont:getHeight()
            )
        )
        buildsavechoices(uistate, cansave)
        gui.layout:pop()
        -- Scroll through the list of save games.
        local presseddown = gui.getPressedKey() == "down"
            or gui.isHit(savedown)
        local pressedup = gui.getPressedKey() == "up"
            or gui.isHit(saveup)
        local current = savestate.newsave or 1
        -- We display the first save file at the top, and scroll down.
        local indexchange =
            ((presseddown and current < #visiblesaves) and 1)
            or ((pressedup and current > 1) and -1)
            or 0
        savestate.oldsave, savestate.newsave = current, current + indexchange
        if savestate.oldsave ~= savestate.newsave then
            savestate.textinput.text = visiblesaves[savestate.newsave]
        end
    end
end

--[[
-- Given the current state of the UI, builds widgits that displays the
-- confirmation text.
--
-- uistate: The current state of the game
--
--]]
local function saveconfirmation(uistate)
    if LOVE then
        local confirmationtext = uistate.savesuccess
            and "'" .. uistate.save.textinput.text .. "'" .. "saved successfully!"
            or uistate.savemessage

        local options = uistate.core.gamestate.options
        local locationtext = "Save location: " .. love.filesystem.getSaveDirectory()
        local titlefont = fonts.title(options.titlefont)
        uistate.gui.Label(
            confirmationtext,
            {id = "confirmation", font = titlefont},
            uistate.gui.layout:row(
                fonts.width(titlefont, locationtext),
                titlefont:getHeight()
            )
        )
        uistate.gui.Label(
            locationtext,
            {id = "location", font = fonts.body(options.bodyfont)},
            uistate.gui.layout:row()
        )

        uistate.gui.Button(
            "(Enter) Continue",
            {font = fonts.button(options.buttonfont), id = "return"},
            uistate.gui.layout:row()
        )
    else
        print(
            string.format(
                "Saved successfully to saves%s%s",
                PATH_SEPARATOR,
                uistate.save.textinput.text
            )
        )
        print("(Enter) Done")
    end
end

--[[
-- Returns a page that saves the game, prints a confirmation, and then
-- transitions back to the game.
--]]
local function buildsavepage()
    local savepage = pages.page(
        nil,
        nil,
        function()
            return {["return"]=uiutilities.backpage(), _keys={"return"}}
        end,
        function(uistate)
            local savename = uistate.save.textinput.text
            if not string.find(savename, ".sav") then
                savename = savename .. ".sav"
            end
            local success, message = gamesaveload.save(
                uistate.core.gamestate,
                savename
            )
            uistate.savesuccess = success
            uistate.savemessage = message
            if not success then
                log.error(message)
            end
            uistate.redrawsave = false
            uistate.visiblesaves = nil
            if uistate.saves then
                local savepath = "saves" .. PATH_SEPARATOR .. savename
                local exists = false
                for _, path in ipairs(uistate.saves) do
                    if path == savepath then
                        exists = true
                        break
                    end
                end
                if not exists then
                    table.insert(uistate.saves, savepath)
                end
            end
        end
    )
    savepage.drawgui = saveconfirmation
    return savepage
end

--[[
-- Given a function that takes the name of a page, and
-- returns a UI story page, returns a page that loads the specified game
-- transitions to the page in the loaded game and clears the save data.
--
-- storypagebuilder: A function that takes the name of a story page returns a
-- UI storypage
--
-- returns: A page that loads the currently selected save game and starts play
-- from that point
--]]
local function buildloadpage(storypagebuilder)
    local loadpage = pages.page(
        nil,
        function(uistate)
            return not uistate.savemessage
                and storypagebuilder(uistate.core.gamestate.storypagename)
        end,
        function()
            return {["return"]=uiutilities.backpage(), _keys={"return"}}
        end,
        function(uistate)
            local gamestate, message = gamesaveload.load(
                uistate.save.textinput.text,
                uistate.core.gamestate.configuration
            )
            uistate.savemessage = message
            if not message then
                uistate.core.gamestate = gamestate
                uistate.line = 1
            end
            uistate.save.allsaves = nil
            uistate.save.oldsave = 1
            uistate.save.newsave = 1
            uistate.save.textinput.text = ""
            uistate.redrawsave = false
            uistate.visiblesaves = nil
        end
    )
    function loadpage.drawgui(uistate)
        local options = uistate.core.gamestate.options
        local titlefont = fonts.title(options.titlefont)
        if uistate.savemessage then
            uistate.gui.Label(
                uistate.savemessage,
                {id="loadmessage", font=fonts.button(options.buttonfont)},
                uistate.gui.layout:row(
                    titlefont:getWidth(uistate.savemessage),
                    titlefont:getHeight()
                )
            )
            uistate.gui.Button(
                "(Enter) Continue",
                {id="return", fonts=fonts.button(options.buttonfont)},
                uistate.gui.layout:row(
                    titlefont:getWidth("(Enter) Continue"),
                    titlefont:getHeight()
                )
            )
        end
    end
    return loadpage
end

--[[
-- Given the current state of the game, returns a table mapping widgit ids
-- to pages. The choices are:
--
-- 1. (F8)Save - Saves the game to the file with the specified name.
-- 2. (F9)Load - Loads the game from file with the specified name.
-- 3. (Esc)Cancel - Return to the game without doing anything
--
-- uistate: The current state of the game
-- ca
--
-- return: A table of the available choices when saving and loading the game
--]]
local function savechoices(cansave)
    local storypages = require 'story/storyui'
    return function()
        local choices = {_keys={}}
        local loadkey = LOVE and "f9" or "l"
        choices[loadkey]=buildloadpage(storypages.page),
        table.insert(choices._keys, loadkey)
        local backkey = LOVE and "escape" or "return"
        table.insert(choices._keys, backkey)
        choices[backkey]=uiutilities.backpage(
            function(uistate1)
                uistate1.redrawsave = false
                uistate1.visiblesaves = nil
            end
        )
        if cansave then
            local savekey = LOVE and "f5" or "s"
            choices[savekey] = buildsavepage()
            table.insert(choices._keys, 1, savekey)
        end
        return choices
   end
end

--[[
-- Returns a list of save files read in from `saves/saves`.
--]]
local function consolesaves()
    local saves = {}
    for line in io.lines("saves" .. PATH_SEPARATOR .. "saves") do
        table.insert(saves, line)
    end
    return saves
end

--[[
-- Builds a page that displays the screen for saving/loading the game
--
-- transformer: The transformer to apply when transitioning to this state
--  (optional)
-- cansave: Whether or not saving should be enabled, defaults to true
--]]
function M.saveload(transformer, cansave)
    if cansave == nil then
        cansave = true
    end
    local page = pages.page(
        nil,
        nil,
        savechoices(cansave),
        function(uistate, previousPage)
            if transformer then transformer(uistate) end
            uiutilities.makePageBackable(uistate, previousPage)
            local saveitems = {}
            if LOVE then 
                local directoryitems = love.filesystem.getDirectoryItems("")
                for _, filename in pairs(saveitems) do
                    if love.filesystem.getInfo(filename, "file") then
                        table.insert(saveitems, filename)
                    end
                end

            else
                if not uistate.saves then
                    uistate.saves = consolesaves()
                end
                local pattern = "saves" .. PATH_SEPARATOR .. "(.+)$"
                for i, filename in ipairs(uistate.saves) do
                    local name = string.match(filename, pattern)
                    saveitems[i] = name
                end
            end
            uistate.allsaves = {}
            for _, filename in pairs(saveitems) do
                log.trace("File name: " .. filename)
                local name, extension = string.match(
                    filename,
                    "^(.+)%.(.+)$"
                )
                if extension == "sav" then
                    log.debug("Found save file: " .. name)
                    table.insert(uistate.allsaves, name)
                end
            end
            table.sort(uistate.allsaves)
            if uistate.save.textinput.text == "" then
                uistate.save.textinput.text = uistate.allsaves[1]
            end
            uistate.command = nil
        end
    )
    page.drawgui = function(uistate) return savegui(uistate, cansave) end
    return page
end

return M
