--[[
-- Copyright 2016 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- titlescreenERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]

--[[
-- Defines the pages that control the Title screen. We have five pages:
--
-- 1. title - Displays the title screen, with three choices: start, load, exit
-- 2. start - Starts a game, by transitioning to the character creation screen.
-- 3. load - Loads a game.
-- 4. Exit - exits the game.
-- 5. confirm - This can be used by any other page that wants to be able to
--  transition to the title. It asks for confirmation before going back to the
--  title screen.
--]]
local M = {}

local os = require 'os'

local love = LOVE and require 'love'

local buttons = LOVE and require 'ui/buttons'
local charactercreationscreen = require 'characters/creationui'
local pages = require 'pages'
local fonts = LOVE and require 'ui/fonts'
local options = require 'utilities/optionsui'
local saveloadui = require 'utilities/saveloadui'
local scarlettables = require 'scarlettables'
local uiutils = require 'ui/utilities'

M.exit = pages.page(
    function() return {} end,
    nil,
    nil,
    function() os.exit(0) end
)


local START_BUTTON_ID = "s"
local function buildStartButton(uistate)
    return buttons.buildButton(
        uistate,
        "s)tart",
        START_BUTTON_ID,
        function ()
            local width = love.graphics.getDimensions()
            return uistate.gui.layout:row(width / 6, 30)
        end
    )
end

local LOAD_BUTTON_ID = "l"
local function buildLoadButton(uistate)
    return buttons.buildButton(uistate, "(l)oad", LOAD_BUTTON_ID, uiutils.col)
end

local EXIT_BUTTON_ID = LOVE and "escape" or "e"
local function buildExitButton(uistate)
    return buttons.buildButton(uistate, "(Esc)Exit", EXIT_BUTTON_ID, uiutils.col)
end

local function buildIncreaseTextButton(uistate)
    return buttons.buildButton(
        uistate,
        "(+) Increase font",
        "+",
        uiutils.col
    )
end

local function buildDecreaseTextButton(uistate)
    return buttons.buildButton(
        uistate,
        "(-) Decrease font",
        "-",
        uiutils.col
    )
end

local function buildToggleFullScreen(uistate)
    return buttons.buildButton(
        uistate,
        "(Ctrl-Enter) Window",
        "",
        uiutils.col
    )
end


M.title = pages.page(
    nil,
    nil,
    function()
        local storypages = require 'story/storyui'
        local startpage = storypages.page("start")
        -- Rather awkward construct that allows us to resolve
        -- a circular dependency between the character creation
        -- and options screens.
        local charactercreation = {}
        local temp = charactercreationscreen.charactercreationscreen(
                options.options(startpage, nil, charactercreation),
                M.title
        )
        for key, uifunction in pairs(temp) do
            charactercreation[key] = uifunction
        end
        return {
            [START_BUTTON_ID] = charactercreation,
            [LOAD_BUTTON_ID] = saveloadui.saveload(nil, false),
            [EXIT_BUTTON_ID] = M.exit,
            ["="]=LOVE and fonts.increasefontsizepage(M.title) or nil,
            ["+"]=LOVE and fonts.increasefontsizepage(M.title, nil, true) or nil,
            ["-"]=LOVE and fonts.decreasefontsizepage(M.title) or nil,
            _keys = {
                START_BUTTON_ID,
                LOAD_BUTTON_ID,
                EXIT_BUTTON_ID,
                "=",
                "-"
            }
        }
    end
)

local titleImage = LOVE and love.graphics.newImage("data/titlescreen.png")
local mask = LOVE 
    and love.image.newImageData(titleImage:getWidth(), titleImage:getHeight())

function M.title.drawgui(uistate)
    if LOVE then
        uistate.gui.layout:push(uistate.gui.layout:padding())
        buildStartButton(uistate)
        buildLoadButton(uistate)
        buildExitButton(uistate)
        buildIncreaseTextButton(uistate)
        buildDecreaseTextButton(uistate)
        buildToggleFullScreen(uistate)
        uistate.gui.layout:pop()
        uistate.gui.ImageButton(
            titleImage,
            {mask=mask},
            uistate.gui.layout:row(0, 0)
        )
    else
        print("------------ The Scandalous Scarlet Moon --------------")
        print(table.concat({"(s)tart", "(l)oad", "(e)xit"}, "\t"))
    end
end


--[[
-- Builds a UI page that asks the player for confirmation before returning
-- to the title
--
-- currentpagename: The name of the current game page
-- previousui: A function that returns the previous UI page (i.e. the page to
--  return to if the player decides not to go back to the title)
--
-- returns A UI page that asks for confirmation before returning to the title
-- page
--]]
function M.confirm(previousuipage)
    local state = require 'state'
    local titlepage = scarlettables.copy(M.title)
    function titlepage.transformer(uistate)
        local newuistate = state.builduistate(
            state.buildgamestate(uistate.core.gamestate.configuration)
        )
        for key in pairs(uistate.core.gamestate) do
            uistate.core.gamestate[key] = newuistate.core.gamestate[key]
        end
    end

    local function choices()
        return {y=titlepage, n=previousuipage(), _keys={"y", "n"}}
    end
    local confirmpage = pages.page(nil, nil, choices)
    function confirmpage.drawgui(uistate)
        if LOVE then
            local bodyfont = fonts.body(
                tonumber(uistate.core.gamestate.options.bodyfont)
            )
            uistate.gui.Label(
                "Are you sure you want to quit?",
                {font = bodyfont},
                uistate.gui.layout:row(300, 30)
            )
            buttons.buildButton(
                uistate,
                "Y)es",
                "y",
                function ()
                    local width = love.graphics.getDimensions()
                    return uistate.gui.layout:row(width / 6)
                end
            )
            buttons.buildButton(
                uistate,
                "N)o",
                "n",
                function ()
                    local width = love.graphics.getDimensions()
                    return uistate.gui.layout:row(width / 6)
                end
            )
        else
            print("Are you sure you want to quit?")
            print("(Y)es\t(N)o")
        end
    end
    return confirmpage
end

return M
